# ----------------------------------------------------------------------------------------
# Routines for libceed
# ----------------------------------------------------------------------------------------
import sys
import os
import re
import subprocess

import build_globals as bglb
from build_consts import *
from build_utils import *

__all__ = ["make_libceed"]

def make_libceed(serial=False):
    if bglb.verbose:
        print("Building libceed")

    path = os.path.join(extdir, 'libceed')
    if not os.path.exists(path):
        assert False, "libceed is not downloaded"

    pwd = chdir(path)
    try:
        make_call(['make', 'clean'])
    except:
        pass

    if bglb.enable_cuda:
        command = ['make', 'configure', 'CUDA_DIR='+bglb.cuda_prefix]
        make_call(command)

    make('libceed')
    make_install('libceed', prefix=bglb.libceed_prefix)
    os.chdir(pwd)



import sys
import os

if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True

    from mfem.common.mpi_debug import nicePrint as print

    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank

else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def test_course_to_file_map1():
    dir = os.path.dirname(os.path.abspath(__file__))        
    mesh_file = os.path.join(dir, "../data/inline-quad.mesh")


    mesh = mfem.Mesh(mesh_file, 1, 1)

    refinements = mfem.RefinementArray()
    refinements.Append(mfem.Refinement(0, 0b11))
    refinements.Append(mfem.Refinement(1, 0b10))
    refinements.Append(mfem.Refinement(2, 0b01))

    mesh.GeneralRefinement(refinements)

    cft = mesh.GetRefinementTransforms()
    coarse_to_fine = mfem.Table()

    cft.MakeCoarseToFineTable(coarse_to_fine)
    print("coarse_to_fine element number mapping:")
    coarse_to_fine.Print()

    '''
    TODO: this part does not work due to API change.
    coarse_to_ref_type = mfem.intArray()
    ref_type_to_matrix = mfem.Table()
    ref_type_to_geom = mfem.GeometryTypeArray()

    cft.GetCoarseToFineMap(mesh,
                           coarse_to_fine,
                           coarse_to_ref_type,
                           ref_type_to_matrix,
                           ref_type_to_geom)

    print("ref_type_to_geom mapping:")

    for i in range(ref_type_to_geom.Size()):
        g = ref_type_to_geom[i]
        print("ref_type: " + str(i) + "=> geom: " + str(g))
   '''

def test_course_to_file_map2():
    dir = os.path.dirname(os.path.abspath(__file__))        
    mesh_file = os.path.join(dir, "../data/inline-quad.mesh")
    
    mesh = mfem.Mesh(mesh_file, 1, 1)

    els = mfem.intArray([0])
    mesh.GeneralRefinement(els)

    zero = mfem.Vector(mesh.GetNE())
    zero.Assign(0.0)

    print(zero.Size())
    mesh.DerefineByError(zero, 1.0)
    print(mesh.GetNE())
    dtrans = mesh.ncmesh.GetDerefinementTransforms()
    coarse_to_fine = mfem.Table()
    
    dtrans.GetCoarseToFineMap(mesh, coarse_to_fine)
    print("table rows: " + str(coarse_to_fine.Size()))
    
    for pe in range(coarse_to_fine.Size()):
        row = mfem.intArray()
        coarse_to_fine.GetRow(pe, row)
        nchild = row.Size();
        txt = "row " + str(pe) + ":"
        for fe in range(nchild):
            child = row[fe]
            txt += " " + str(child)
        print(txt)

if __name__ == '__main__':
    device = mfem.Device('cpu')
    
    test_course_to_file_map1()
    test_course_to_file_map2()




"""
Helper functions for setup.py
"""

import os
import sys
import configparser
from urllib import request
import itertools
import site
import re
import subprocess
import multiprocessing
import ssl
import tarfile
import shutil
from collections import namedtuple
from shutil import which as find_command

__all__ = ["print_config",
           "initialize_cmd_options",
           "cmd_options",
           "process_cmd_options",
           "configure_build",
           "clean_dist_info",
           ]

from build_utils import *
from build_consts import *
import build_globals as bglb


def print_config():
    print("----configuration----")
    print(" prefix", bglb.prefix)
    print(" when needed, the dependency (mfem/hypre/metis) will be installed under " +
          bglb.ext_prefix)
    print(" build mfem : " + ("Yes" if bglb.build_mfem else "No"))
    print(" build metis : " + ("Yes" if bglb.build_metis else "No"))
    print(" build hypre : " + ("Yes" if bglb.build_hypre else "No"))
    print(" build libceed : " + ("Yes" if bglb.build_libceed else "No"))
    print(" build gslib : " + ("Yes" if bglb.build_gslib else "No"))
    print(" call SWIG wrapper generator: " +
          ("Yes" if bglb.run_swig else "No"))
    print(" build serial wrapper: " + ("Yes" if bglb.build_serial else "No"))
    print(" build parallel wrapper : " +
          ("Yes" if bglb.build_parallel else "No"))

    print(" hypre prefix", bglb.hypre_prefix)
    print(" metis prefix", bglb.metis_prefix)
    print(" c compiler : " + bglb.cc_command)
    print(" c++ compiler : " + bglb.cxx_command)
    print(" mpi-c compiler : " + bglb.mpicc_command)
    print(" mpi-c++ compiler : " + bglb.mpicxx_command)

    print(" verbose : " + ("Yes" if bglb.verbose else "No"))
    print(" SWIG : " + swig_command)

    if bglb.blas_libraries != "":
        print(" BLAS libraries : " + bglb.blas_libraries)
    if bglb.lapack_libraries != "":
        print(" Lapack libraries : " + bglb.lapack_libraries)

    print("")


def clean_dist_info(wheeldir):
    if not os.path.isdir(wheeldir):
        return
    for x in os.listdir(wheeldir):
        if x.endswith(".dist-info"):
            fname = os.path.join(wheeldir, x)
            print("!!! removing existing ", fname)
            shutil.rmtree(fname)


def initialize_cmd_options(command_obj):
    command_obj.swig = False
    command_obj.skip_swig = False
    command_obj.ext_only = False

    command_obj.git_sshclone = False
    command_obj.skip_ext = False
    command_obj.with_parallel = False
    command_obj.no_serial = False
    command_obj.mfem_prefix = ''
    command_obj.mfems_prefix = ''
    command_obj.mfemp_prefix = ''
    command_obj.mfem_source = bglb.mfem_source
    command_obj.mfem_branch = ''
    command_obj.mfem_debug = False
    command_obj.mfem_build_miniapps = False
    command_obj.metis_prefix = ''
    command_obj.hypre_prefix = ''

    command_obj.with_cuda = False
    command_obj.with_cuda_hypre = False
    command_obj.cuda_arch = None
    command_obj.with_metis64 = False

    command_obj.with_pumi = False
    command_obj.pumi_prefix = ''

    command_obj.with_strumpack = False
    command_obj.strumpack_prefix = ''

    command_obj.with_suitesparse = False
    command_obj.suitesparse_prefix = ''

    command_obj.with_lapack = False
    command_obj.blas_libraries = ""
    command_obj.lapack_libraries = ""

    command_obj.with_libceed = False
    command_obj.libceed_prefix = ''
    command_obj.libceed_only = False

    command_obj.with_gslib = False
    command_obj.gslib_prefix = ''
    command_obj.gslib_only = False

    command_obj.CC = ''
    command_obj.CXX = ''
    command_obj.MPICC = ''
    command_obj.MPICXX = ''
    command_obj.vv = False

    command_obj.unverifiedSSL = False


cmd_options = [
    ('vv', None, 'More verbose output (CMAKE_VERBOSE_MAKEFILE etc)'),
    ('with-parallel', None, 'Installed both serial and parallel version'),
    ('no-serial', None, 'Skip building the serial wrapper'),
    ('mfem-prefix=', None, 'Specify locaiton of mfem' +
     'libmfem.so must exits under <mfem-prefix>/lib. ' +
     'This mode uses clean-swig + run-swig, unless mfem-prefix-no-swig is on'),
    ('mfemp-prefix=', None, 'Specify locaiton of parallel mfem ' +
     'libmfem.so must exits under <mfemp-prefix>/lib. ' +
     'Need to use it with mfem-prefix'),
    ('mfems-prefix=', None, 'Specify locaiton of serial mfem ' +
     'libmfem.so must exits under <mfems-prefix>/lib. ' +
     'Need to use it with mfem-prefix'),
    ('mfem-branch=', None, 'Specify branch of mfem' +
     'MFEM is cloned and built using the specfied branch '),
    ('mfem-source=', None, 'Specify mfem source location' +
     'MFEM source directory. Required to run-swig '),
    ('mfem-debug', None, 'Build MFME with MFEM_DEBUG enabled'),
    ('mfem-build-miniapps', None, 'build MFME Miniapps'),
    ('hypre-prefix=', None, 'Specify locaiton of hypre' +
     'libHYPRE.so must exits under <hypre-prefix>/lib'),
    ('metis-prefix=', None, 'Specify locaiton of metis' +
     'libmetis.so must exits under <metis-prefix>/lib'),
    ('git-sshclone', None, 'Use SSH for git clone' +
     'try if default git clone using https fails (need Github account and setting for SSH)'),
    ('swig', None, 'Run Swig and exit'),
    ('skip-swig', None,
     'Skip running swig (used when wrapper is generated for the MFEM C++ library to be used'),
    ('ext-only', None, 'Build metis, hypre, mfem(C++) only'),
    ('skip-ext', None, 'Skip building metis, hypre, mfem(C++) only'),
    ('CC=', None, 'c compiler'),
    ('CXX=', None, 'c++ compiler'),
    ('MPICC=', None, 'mpic compiler'),
    ('MPICXX=', None, 'mpic++ compiler'),
    ('unverifiedSSL', None, 'use unverified SSL context for downloading'),
    ('with-cuda', None, 'enable cuda'),
    ('with-cuda-hypre', None, 'enable cuda in hypre'),
    ('cuda-arch=', None, 'set cuda compute capability. Ex if A100, set to 80'),
    ('with-metis64', None, 'use 64bit int in metis'),
    ('with-pumi', None, 'enable pumi (parallel only)'),
    ('pumi-prefix=', None, 'Specify locaiton of pumi'),
    ('with-suitesparse', None,
     'build MFEM with suitesparse (MFEM_USE_SUITESPARSE=YES) (parallel only)'),
    ('suitesparse-prefix=', None,
     'Specify locaiton of suitesparse (=SuiteSparse_DIR)'),
    ('with-libceed', None, 'enable libceed'),
    ('libceed-prefix=', None, 'Specify locaiton of libceed'),
    ('libceed-only', None, 'Build libceed only'),
    ('gslib-prefix=', None, 'Specify locaiton of gslib'),
    ('with-gslib', None, 'enable gslib'),
    ('gslib-only', None, 'Build gslib only'),
    ('with-strumpack', None, 'enable strumpack (parallel only)'),
    ('strumpack-prefix=', None, 'Specify locaiton of strumpack'),
    ('with-lapack', None, 'build MFEM with lapack'),
    ('blas-libraries=', None, 'Specify locaiton of Blas library (used to build MFEM)'),
    ('lapack-libraries=', None,
     'Specify locaiton of Lapack library (used to build MFEM)'),
]


def process_cmd_options(command_obj, cfs):
    '''
    called when install workflow is used
    '''
    cc = cfs.pop("CC", "")
    if cc != "":
        command_obj.cc_command = cc

    cc = cfs.pop("CXX", "")
    if cc != "":
        command_obj.cxx_command = cc

    cc = cfs.pop("MPICC", "")
    if cc != "":
        command_obj.mpicc_command = cc

    cc = cfs.pop("MPICXX", "")
    if cc != "":
        command_obj.mpicxx_command = cc

    for item in cmd_options:
        param, _none, hit = item
        attr = "_".join(param.split("-"))

        if param.endswith("="):
            param = param[:-1]
            attr = attr[:-1]
            value = cfs.pop(param, "")
            if value != "":
                if not hasattr(command_obj, attr):
                    assert False, str(command_obj) + " does not have " + attr
                setattr(command_obj, attr, value)
        else:
            value = cfs.pop(param, "No")
            if not hasattr(command_obj, attr):
                assert False, str(command_obj) + " does not have " + attr

            if value.upper() in ("YES", "TRUE", "1"):
                setattr(command_obj, attr, True)
            else:
                setattr(command_obj, attr, False)


def process_setup_options(command_obj, args):
    for item in args:
        if item.startswith('--'):
            item = item[2:]
        if item.startswith('-'):
            item = item[1:]

        if len(item.split('=')) == 2:
            param = item.split('=')[0]
            value = item.split('=')[1]
        else:
            param = item.strip()
            value = True
        attr = "_".join(param.split("-"))

        setattr(command_obj, attr, value)


def configure_install(self):
    '''
    called when install workflow is used

    '''
    if sys.argv[0] == 'setup.py' and sys.argv[1] == 'install':
        process_setup_options(self, sys.argv[2:])
    else:
        if bglb.verbose:
            print("!!!!!!!!  command-line input (pip): ", bglb.cfs)
        process_cmd_options(self, bglb.cfs)

    bglb.verbose = bool(self.vv) if not bglb.verbose else bglb.verbose
    if bglb.dry_run:
        bglb.verbose = True

    bglb.git_sshclone = bool(self.git_sshclone)

    bglb.mfem_source = abspath(self.mfem_source)

    bglb.skip_ext = bool(self.skip_ext)
    bglb.skip_swig = bool(self.skip_swig)

    bglb.swig_only = bool(self.swig)
    bglb.ext_only = bool(self.ext_only)

    bglb.metis_64 = bool(self.with_metis64)
    bglb.enable_pumi = bool(self.with_pumi)
    bglb.enable_strumpack = bool(self.with_strumpack)
    bglb.enable_cuda = bool(self.with_cuda)
    bglb.enable_cuda_hypre = bool(self.with_cuda_hypre)
    if self.cuda_arch is not None:
        bglb.cuda_arch = self.cuda_arch
    bglb.enable_libceed = bool(self.with_libceed)
    bglb.libceed_only = bool(self.libceed_only)
    bglb.enable_gslib = bool(self.with_gslib)
    bglb.gslib_only = bool(self.gslib_only)
    bglb.enable_suitesparse = bool(self.with_suitesparse)
    bglb.enable_lapack = bool(self.with_lapack)

    # controlls PyMFEM parallel
    bglb.build_parallel = bool(self.with_parallel)
    bglb.build_serial = not bool(self.no_serial)

    bglb.clean_swig = True
    bglb.run_swig = True
    bglb.run_swig_parallel = bool(self.with_parallel)

    bglb.mfem_debug = bool(self.mfem_debug)
    bglb.mfem_build_miniapps = bool(self.mfem_build_miniapps)

    if bglb.build_serial:
        bglb.build_serial = (not bglb.swig_only and not bglb.ext_only)

    if bglb.build_parallel:
        try:
            import mpi4py
        except ImportError:
            assert False, "Can not import mpi4py"

    if self.mfem_prefix != '':
        bglb.mfem_prefix = abspath(self.mfem_prefix)
        bglb.mfems_prefix = abspath(self.mfem_prefix)
        bglb.mfemp_prefix = abspath(self.mfem_prefix)
        if self.mfems_prefix != '':
            bglb.mfems_prefix = abspath(self.mfems_prefix)
        if self.mfemp_prefix != '':
            bglb.mfemp_prefix = abspath(self.mfemp_prefix)

        check = find_libpath_from_prefix('mfem', bglb.mfems_prefix)
        assert check != '', "libmfem.so is not found in the specified <path>/lib"
        check = find_libpath_from_prefix('mfem', bglb.mfemp_prefix)
        assert check != '', "libmfem.so is not found in the specified <path>/lib"

        bglb.mfem_outside = True
        bglb.build_mfem = False
        hypre_prefix = bglb.mfem_prefix
        metis_prefix = bglb.mfem_prefix

        if bglb.swig_only:
            bglb.clean_swig = False

    else:
        bglb.mfem_outside = False
        bglb.build_mfem = True
        bglb.build_mfemp = bglb.build_parallel
        bglb.build_hypre = bglb.build_parallel
        bglb.build_metis = bglb.build_parallel or bglb.enable_suitesparse

        if bglb.ext_prefix == '':
            bglb.ext_prefix = external_install_prefix(bglb.prefix)
        bglb.hypre_prefix = os.path.join(bglb.ext_prefix)
        bglb.metis_prefix = os.path.join(bglb.ext_prefix)

        bglb.mfem_prefix = bglb.ext_prefix
        bglb.mfems_prefix = os.path.join(bglb.ext_prefix, 'ser')
        bglb.mfemp_prefix = os.path.join(bglb.ext_prefix, 'par')
        # enable_gslib = True

    if self.mfem_branch != '':
        bglb.mfem_branch = self.mfem_branch

    if self.hypre_prefix != '':
        check = find_libpath_from_prefix('HYPRE', self.hypre_prefix)
        assert check != '', "libHYPRE.so is not found in the specified <path>/lib or lib64"
        hypre_prefix = os.path.expanduser(self.hypre_prefix)
        build_hypre = False

    if self.metis_prefix != '':
        check = find_libpath_from_prefix('metis', self.metis_prefix)
        assert check != '', "libmetis.so is not found in the specified <path>/lib or lib64"
        bglb.metis_prefix = os.path.expanduser(self.metis_prefix)
        bglb.build_metis = False

    if bglb.enable_libceed or bglb.libceed_only:
        if self.libceed_prefix != '':
            bglb.libceed_prefix = os.path.expanduser(self.libceed_prefix)
            bglb.build_libceed = False
        else:
            bglb.libceed_prefix = bglb.mfem_prefix
            bglb.build_libceed = True

    if bglb.enable_gslib or bglb.gslib_only:
        if self.gslib_prefix != '':
            bglb.build_gslib = False
            bglb.gslibs_prefix = os.path.expanduser(self.gslib_prefix)
            bglb.gslibp_prefix = os.path.expanduser(self.gslib_prefix)
        else:
            bglb.gslibs_prefix = bglb.mfems_prefix
            bglb.gslibp_prefix = bglb.mfemp_prefix
            bglb.build_gslib = True

    if bglb.enable_suitesparse and self.suitesparse_prefix != '':
        bglb.suitesparse_prefix = self.suitesparse_prefix

    if self.pumi_prefix != '':
        bglb.pumi_prefix = abspath(self.pumi_prefix)
    else:
        bglb.pumi_prefix = bglb.mfem_prefix

    if self.strumpack_prefix != '':
        bglb.strumpack_prefix = abspath(self.strumpack_prefix)
    else:
        bglb.strumpack_prefix = bglb.mfem_prefix

    if bglb.enable_cuda:
        nvcc = find_command('nvcc')
        bglb.cuda_prefix = os.path.dirname(os.path.dirname(nvcc))

    if self.CC != '':
        bglb.cc_command = self.CC
    if self.CXX != '':
        bglb.cxx_command = self.CXX
    if self.MPICC != '':
        bglb.mpicc_command = self.MPICC
    if self.MPICXX != '':
        bglb.mpicxx_command = self.MPICXX

    if self.blas_libraries != "":
        bglb.blas_libraries = self.blas_libraries
    if self.lapack_libraries != "":
        bglb.lapack_libraries = self.lapack_libraries

    if bglb.swig_only:
        bglb.build_serial = False
        bglb.build_parallel = False
        bglb.clean_swig = False
        bglb.keep_temp = True
        bglb.skip_ext = True

    if bglb.skip_ext:
        bglb.build_metis = False
        bglb.build_hypre = False
        bglb.build_mfem = False
        bglb.build_mfemp = False
        bglb.build_libceed = False
        bglb.build_gslib = False

    if bglb.skip_swig:
        bglb.clean_swig = False
        bglb.run_swig = False

    if bglb.ext_only:
        bglb.clean_swig = False
        bglb.run_swig = False
        bglb.build_serial = False
        bglb.build_parallel = False
        bglb.keep_temp = True


    if bglb.libceed_only:
        bglb.clean_swig = False
        bglb.run_swig = False
        bglb.build_mfem = False
        bglb.build_mfemp = False
        bglb.build_metis = False
        bglb.build_hypre = False
        bglb.build_gslib = False
        bglb.build_serial = False
        bglb.build_parallel = False
        bglb.build_libceed = True
        bglb.keep_temp = True


    if bglb.gslib_only:
        bglb.clean_swig = False
        bglb.run_swig = False
        bglb.build_mfem = False
        bglb.build_mfemp = False
        bglb.build_metis = False
        bglb.build_hypre = False
        bglb.build_serial = False
        bglb.build_libceed = False
        bglb.build_gslib = True
        bglb.keep_temp = True


    bglb.is_configured = True

configure_build = configure_install



'''
   Matrix coefficient example

   How to run:
      python <arguments>

   Example of arguments:
      coefficient_example.py

   Description:  This example code demonstrates how to use matrix coefficient,
                 which is defiend in Python.

'''
import mfem.ser as mfem
import numpy as np

class sigma(mfem.MatrixPyCoefficient):
    def __init__(self, dim):
       mfem.MatrixPyCoefficient.__init__(self, dim)
    def EvalValue(self, x):
       return np.array([[0.1, 0.], [0., 10.]])
   
# create sample mesh for square shape
mesh = mfem.Mesh(10, 10, "TRIANGLE")

# create finite element function space
fec = mfem.H1_FECollection(1, mesh.Dimension())   # H1 order=1
fespace = mfem.FiniteElementSpace(mesh, fec)      

# 
ess_tdof_list = mfem.intArray()
ess_bdr = mfem.intArray([1]*mesh.bdr_attributes.Size())
fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

# constant coefficient 
one = mfem.ConstantCoefficient(1.0)

# define Bilinear and Linear operator
a = mfem.BilinearForm(fespace)
a.AddDomainIntegrator(mfem.DiffusionIntegrator(sigma(2)))
a.Assemble()
b = mfem.LinearForm(fespace)
b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
b.Assemble()

# create gridfunction, which is where the solution vector is stored
x = mfem.GridFunction(fespace);
x.Assign(0.0)

# form linear equation (AX=B)
A = mfem.OperatorPtr()
B = mfem.Vector()
X = mfem.Vector()
a.FormLinearSystem(ess_tdof_list, x, b, A, X, B);
print("Size of linear system: " + str(A.Height()))

# solve it using PCG solver and store the solution to x
AA = mfem.OperatorHandle2SparseMatrix(A)
M = mfem.GSSmoother(AA)
mfem.PCG(AA, M, B, X, 1, 200, 1e-12, 0.0)
a.RecoverFEMSolution(X, b, x)

# extract vertices and solution as numpy array
verts = mesh.GetVertexArray()
sol = x.GetDataArray()

# plot solution using Matplotlib

import matplotlib.pyplot as plt
import matplotlib.tri as tri

triang = tri.Triangulation(verts[:,0], verts[:,1])

fig1, ax1 = plt.subplots()
ax1.set_aspect('equal')
tpc = ax1.tripcolor(triang, sol, shading='gouraud')
fig1.colorbar(tpc)
plt.show()



from argparse import ArgumentParser

class ArgParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        self.argument_list = []
        ArgumentParser.__init__(self, *args, **kwargs)


    def add_argument(self, *args, **kwargs):
        ArgumentParser.add_argument(self, *args, **kwargs)
        self.argument_list.append(args[-1])

    def print_options(self, args):
        print("Options used:")
        d = vars(args)

        keys = ['_'.join(filter(None, x.split('-'))) 
                for x in self.argument_list]
        for k in keys:
            if k in d: print("   --" +  k + "  " + str(d[k]))




from __future__ import print_function
import os
import sys

def run_test():
    print("Test segment module")        
    seg = mfem.Segment()
    seg.SetVertices((1,2))
    seg.SetAttribute(1)
    assert  seg.GetVertices() == (1,2), "failed to set attribute (must be (1,2))"
    assert  seg.GetAttribute() == 1, "failed to set attribute (must be 1)"

    seg = mfem.Segment()
    seg.SetVertices((1,2))
    seg.SetAttribute(1)
    assert  seg.GetVertices() == (1,2), "failed to set attribute (must be (1,2))"
    assert  seg.GetAttribute() == 1, "failed to set attribute (must be 1)"

    seg = mfem.Segment([1, 2], 1)
    assert  seg.GetVertices() == (1,2), "failed to set attribute (must be (1,2))"
    assert  seg.GetAttribute() == 1, "failed to set attribute (must be 1)"

    seg = mfem.Segment((1, 2), 1)
    assert  seg.GetVertices() == (1,2), "failed to set attribute (must be (1,2))"
    assert  seg.GetAttribute() == 1, "failed to set attribute (must be 1)"

if __name__=='__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-p':   
        import mfem.par as mfem
    else:
        import mfem.ser as mfem
        
    run_test()

    






'''
   MFEM example 1 (converted from ex1.cpp)

   See c++ version in the MFEM library for more detail

   How to run:
      python <arguments>

   Example of arguments:
      ex1.py -m star.mesh
      ex1.py -m square-disc.mesh
      ex1.py -m escher.mesh
      ex1.py -m fichera.mesh
      ex1.py -m square-disc-p3.mesh -o 3
      ex1.py -m square-disc-nurbs.mesh -o -1
      ex1.py -m disc-nurbs.mesh -o -1
      ex1.py -m pipe-nurbs.mesh -o -1
      ex1.py -m star-surf.mesh
      ex1.py -m square-disc-surf.mesh
      ex1.py -m inline-segment.mesh
      ex1.py -m amr-quad.mesh
      ex1.py -m amr-hex.mesh
      ex1.py -m fichera-amr.mesh
      ex1.py -m mobius-strip.mesh
      ex1.py -m mobius-strip.mesh -o -1 -sc

   Description:  This example code demonstrates the use of MFEM to define a
                 simple finite element discretization of the Laplace problem
                 -Delta u = 1 with homogeneous Dirichlet boundary conditions.

'''
import os
from os.path import expanduser, join
import numpy as np

import mfem.ser as mfem


def run(order=1, static_cond=False,
        meshfile='', visualization=False,
        device='cpu', pa=False):
    '''
    run ex1
    '''
    device = mfem.Device(device)
    device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    #   3. Refine the mesh to increase the resolution. In this example we do
    #      'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
    #      largest number that gives a final mesh with no more than 50,000
    #      elements.
    ref_levels = int(
        np.floor(
            np.log(
                50000. /
                mesh.GetNE()) /
            np.log(2.) /
            dim))
    for x in range(ref_levels):
        mesh.UniformRefinement()

    # 5. Define a finite element space on the mesh. Here we use vector finite
    #   elements, i.e. dim copies of a scalar finite element space. The vector
    #   dimension is specified by the last argument of the FiniteElementSpace
    #   constructor. For NURBS meshes, we use the (degree elevated) NURBS space
    #   associated with the mesh nodes.
    if order > 0:
        fec = mfem.H1_FECollection(order, dim)
    elif mesh.GetNodes():
        fec = mesh.GetNodes().OwnFEC()
        print("Using isoparametric FEs: " + str(fec.Name()))
    else:
        order = 1
        fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec)
    print('Number of finite element unknowns: ' +
          str(fespace.GetTrueVSize()))
    # 5. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined by marking all
    #    the boundary attributes from the mesh as essential (Dirichlet) and
    #    converting them to a list of true dofs.
    ess_tdof_list = mfem.intArray()
    if mesh.bdr_attributes.Size() > 0:
        ess_bdr = mfem.intArray([1] * mesh.bdr_attributes.Max())
        ess_bdr = mfem.intArray(mesh.bdr_attributes.Max())
        ess_bdr.Assign(1)
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)
    # 6. Set up the linear form b(.) which corresponds to the right-hand side of
    #   the FEM linear system, which in this case is (1,phi_i) where phi_i are
    #   the basis functions in the finite element fespace.
    b = mfem.LinearForm(fespace)
    one = mfem.ConstantCoefficient(1.0)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
    b.Assemble()
    # 7. Define the solution vector x as a finite element grid function
    #   corresponding to fespace. Initialize x with initial guess of zero,
    #   which satisfies the boundary conditions.
    x = mfem.GridFunction(fespace)
    x.Assign(0.0)
    # 8. Set up the bilinear form a(.,.) on the finite element space
    #   corresponding to the Laplacian operator -Delta, by adding the Diffusion
    #   domain integrator.
    a = mfem.BilinearForm(fespace)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
    # 9. Assemble the bilinear form and the corresponding linear system,
    #   applying any necessary transformations such as: eliminating boundary
    #   conditions, applying conforming constraints for non-conforming AMR,
    #   static condensation, etc.
    if static_cond:
        a.EnableStaticCondensation()
    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()

    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
    print("Size of linear system: " + str(A.Height()))

    # 10. Solve

    if pa:
        if mfem.UsesTensorBasis(fespace):
            M = mfem.OperatorJacobiSmoother(a, ess_tdof_list)
            mfem.PCG(A, M, B, X, 1, 4000, 1e-12, 0.0)
        else:
            mfem.CG(A, B, X, 1, 400, 1e-12, 0.0)
    else:
        #AA = mfem.OperatorHandle2SparseMatrix(A)
        AA = A.AsSparseMatrix()
        M = mfem.GSSmoother(AA)
        mfem.PCG(A, M, B, X, 1, 200, 1e-12, 0.0)

    # 11. Recover the solution as a finite element grid function.
    a.RecoverFEMSolution(X, b, x)
    # 12. Save the refined mesh and the solution. This output can be viewed later
    #     using GLVis: "glvis -m refined.mesh -g sol.gf".
    mesh.Print('refined.mesh', 8)
    x.Save('sol.gf', 8)

    # 13. Send the solution by socket to a GLVis server.
    if (visualization):
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock.send_solution(mesh, x)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex1 (Laplace Problem)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    args = parser.parse_args()
    parser.print_options(args)

    order = args.order
    static_cond = args.static_condensation

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly

    run(order=order,
        static_cond=static_cond,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa)



'''
   MFEM example 31

   See c++ version in the MFEM library for more detail

   Sample runs:  python ex31.py -m ../data/inline-segment.mesh -o 2
                 python ex31.py -m ../data/hexagon.mesh -o 2
                 python ex31.py -m ../data/star.mesh -o 2
                 python ex31.py -m ../data/fichera.mesh -o 3 -r 1
                 python ex31.py -m ../data/square-disc-nurbs.mesh -o 3
                 python ex31.py -m ../data/amr-quad.mesh -o 2 -r 1
                 python ex31.py -m ../data/amr-hex.mesh -r 1
'''
import mfem.ser as mfem
from mfem.ser import intArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt

sqrt1_2 = 1/sqrt(2)
sqrt2 = sqrt(2)


def run(order=1,
        refine=2,
        freq=1,
        meshfile='',
        visualization=False,
        numba=False):

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    # 3. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement (2 by default, or specified on
    #    the command line with -r)
    for x in range(refine):
        mesh.UniformRefinement()

    # 4. Define a finite element space on the mesh. Here we use the Nedelec
    #    finite elements of the specified order restricted to 1D, 2D, or 3D
    #    depending on the dimension of the given mesh file.
    if dim == 1:
        fec = mfem.ND_R1D_FECollection(order, dim)
    elif dim == 2:
        fec = mfem.ND_R2D_FECollection(order, dim)
    else:
        fec = mfem.ND_FECollection(order, dim)

    fespace = mfem.FiniteElementSpace(mesh, fec)
    print("Number of H(curl) unknowns: " + str(fespace.GetTrueVSize()))

    # 5. Determine the list of true essential boundary dofs. In this example,
    #    the boundary conditions are defined by marking all the boundary
    #    attributes from the mesh as essential (Dirichlet) and converting them
    #    to a list of true dofs.
    ess_tdof_list = intArray()
    if mesh.bdr_attributes.Size():
        ess_bdr = intArray([1]*mesh.bdr_attributes.Max())
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 5.b Define coefficent to use. Here, we define outside this function.
    #     We call decoratr function (mfem.jit.xxx) manually. Also, note
    #     we are passing dim as params, so that numba uses proper dim
    #     when compiling it.
    if numba:
        params = {"dim": dim, "kappa": pi*freq}
        E = mfem.jit.vector(vdim=3, params=params)(E_exact)
        CurlE = mfem.jit.vector(vdim=3, params=params)(CurlE_exact)
        f = mfem.jit.vector(vdim=3, params=params)(f_exact)
    else:
        pass  # ToDo provide regular example.

    # 6. Set up the linear form b(.) which corresponds to the right-hand side
    #    of the FEM linear system, which in this case is (f,phi_i) where f is
    #    given by the function f_exact and phi_i are the basis functions in
    #    the finite element fespace.

    b = mfem.LinearForm(fespace)
    b.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(f))
    b.Assemble()

    # 7. Define the solution vector x as a finite element grid function
    #    corresponding to fespace. Initialize x by projecting the exact
    #    solution. Note that only values from the boundary edges will be used
    #    when eliminating the non-homogeneous boundary condition to modify the
    #    r.h.s. vector b.
    sol = mfem.GridFunction(fespace)
    sol.ProjectCoefficient(E)

    # 8. Set up the bilinear form corresponding to the EM diffusion
    #    operator curl muinv curl + sigma I, by adding the curl-curl and the
    #    mass domain integrators.
    mat = array([[2.0, sqrt1_2, 0, ],
                 [sqrt1_2, 2.0, sqrt1_2],
                 [0.0, sqrt1_2, 2.0, ], ])
    sigmaMat = mfem.DenseMatrix(mat)

    muinv = mfem.ConstantCoefficient(1.0)
    sigma = mfem. MatrixConstantCoefficient(sigmaMat)
    a = mfem.BilinearForm(fespace)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(muinv))
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(sigma))

    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, sol, b, A, X, B)

    # 10. Solve the system A X = B.
    AM = A.AsSparseMatrix()
    M = mfem.GSSmoother(AM)
    mfem.PCG(A, M, B, X, 1, 500, 1e-12, 0.0)

    # 12. Recover the solution as a finite element grid function.
    a.RecoverFEMSolution(X, b, sol)

    # 13. Compute and print the CurlE norm of the error.
    print("|| E_h - E ||_{Hcurl} = " +
          "{:g}".format(sol.ComputeHCurlError(E, CurlE))+"\n")

    # 14. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m refined.mesh -g sol.gf"
    mesh.Print('refined.mesh', 8)
    sol.Save('sol.gf', 8)

    # 15. Send the solution by socket to a GLVis server.
    if visualization:

        solCoef = mfem.VectorGridFunctionCoefficient(sol)
        dsolCoef = mfem.CurlGridFunctionCoefficient(sol)

        if dim == 1:
            x_sock = mfem.socketstream("localhost", 19916)
            y_sock = mfem.socketstream("localhost", 19916)
            z_sock = mfem.socketstream("localhost", 19916)
            dy_sock = mfem.socketstream("localhost", 19916)
            dz_sock = mfem.socketstream("localhost", 19916)

            x_sock.precision(8)
            y_sock.precision(8)
            z_sock.precision(8)
            dy_sock.precision(8)
            dz_sock.precision(8)

            xVec = mfem.Vector([1, 0, 0.])
            yVec = mfem.Vector([0, 1, 0.])
            zVec = mfem.Vector([0, 0, 1.])

            xVecCoef = mfem.VectorConstantCoefficient(xVec)
            yVecCoef = mfem.VectorConstantCoefficient(yVec)
            zVecCoef = mfem.VectorConstantCoefficient(zVec)

            fec_h1 = mfem.H1_FECollection(order, dim)
            fec_l2 = mfem.L2_FECollection(order-1, dim)

            fes_h1 = mfem.FiniteElementSpace(mesh, fec_h1)
            fes_l2 = mfem.FiniteElementSpace(mesh, fec_l2)

            xComp = mfem.GridFunction(fes_l2)
            yComp = mfem.GridFunction(fes_h1)
            zComp = mfem.GridFunction(fes_h1)

            dyComp = mfem.GridFunction(fes_l2)
            dzComp = mfem.GridFunction(fes_l2)

            xCoef = mfem.InnerProductCoefficient(xVecCoef, solCoef)
            yCoef = mfem.InnerProductCoefficient(yVecCoef, solCoef)
            zCoef = mfem.InnerProductCoefficient(zVecCoef, solCoef)

            xComp.ProjectCoefficient(xCoef)
            yComp.ProjectCoefficient(yCoef)
            zComp.ProjectCoefficient(zCoef)

            x_sock << "solution\n" << mesh << xComp
            x_sock.flush()
            x_sock << "window_title 'X component'"
            x_sock.flush()
            y_sock << "solution\n" << mesh << yComp
            y_sock.flush()
            y_sock << "window_geometry 403 0 400 350 " << "window_title 'Y component'"
            y_sock.flush()
            z_sock << "solution\n" << mesh << zComp
            z_sock.flush()
            z_sock << "window_geometry 806 0 400 350 " << "window_title 'Z component'"
            z_sock.flush()

            dyCoef = mfem.InnerProductCoefficient(yVecCoef, dsolCoef)
            dzCoef = mfem.InnerProductCoefficient(zVecCoef, dsolCoef)
            dyComp.ProjectCoefficient(dyCoef)
            dzComp.ProjectCoefficient(dzCoef)

            dy_sock << "solution\n" << mesh << dyComp
            dy_sock.flush()
            dy_sock << "window_geometry 403 375 400 350 " << "window_title 'Y component of Curl'"
            dy_sock.flush()

            dy_sock << "solution\n" << mesh << dzComp
            dy_sock.flush()
            dy_sock << "window_geometry 403 375 400 350 " << "window_title 'Z component of Curl'"
            dy_sock.flush()

        elif dim == 2:
            xy_sock = mfem.socketstream("localhost", 19916)
            z_sock = mfem.socketstream("localhost", 19916)
            dxy_sock = mfem.socketstream("localhost", 19916)
            dz_sock = mfem.socketstream("localhost", 19916)

            mat = array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
            xyMat = mfem.DenseMatrix(mat)
            xyMatCoef = mfem.MatrixConstantCoefficient(xyMat)
            zVec = array([0, 0, 1])
            zVecCoef = mfem.VectorConstantCoefficient(zVec)

            xyCoef = mfem.MatrixVectorProductCoefficient(xyMatCoef, solCoef)
            zCoef = mfem.InnerProductCoefficient(zVecCoef, solCoef)

            fec_h1 = mfem.H1_FECollection(order, dim)
            fec_nd = mfem.ND_FECollection(order, dim)
            fec_rt = mfem.RT_FECollection(order-1, dim)
            fec_l2 = mfem.L2_FECollection(order-1, dim)

            fes_h1 = mfem.FiniteElementSpace(mesh, fec_h1)
            fes_nd = mfem.FiniteElementSpace(mesh, fec_nd)
            fes_rt = mfem.FiniteElementSpace(mesh, fec_rt)
            fes_l2 = mfem.FiniteElementSpace(mesh, fec_l2)

            xyComp = mfem.GridFunction(fes_nd)
            zComp = mfem.GridFunction(fes_h1)

            dxyComp = mfem.GridFunction(fes_rt)
            dzComp = mfem.GridFunction(fes_l2)

            xyComp.ProjectCoefficient(xyCoef)
            zComp.ProjectCoefficient(zCoef)

            xy_sock.precision(8)
            xy_sock << "solution\n" << mesh << xyComp
            xy_sock << "window_title 'XY components'\n"
            xy_sock.flush()
            z_sock << "solution\n" << mesh << zComp
            z_sock << "window_geometry 403 0 400 350 " << "window_title 'Z component'"
            z_sock.flush()

            dxyCoef = mfem.MatrixVectorProductCoefficient(xyMatCoef, dsolCoef)
            dzCoef = mfem.InnerProductCoefficient(zVecCoef, dsolCoef)

            dxyComp.ProjectCoefficient(dxyCoef)
            dzComp.ProjectCoefficient(dzCoef)

            dxy_sock << "solution\n" << mesh << dxyComp
            dxy_sock << "window_geometry 0 375 400 350 " << "window_title 'XY components of Curl'"
            dxy_sock.flush()
            dz_sock << "solution\n" << mesh << dzComp
            dz_sock << "window_geometry 403 375 400 350 " << "window_title 'Z component of Curl'"
            dz_sock.flush()

        else:
            sol_sock = mfem.socketstream("localhost", 19916)
            dsol_sock = mfem.socketstream("localhost", 19916)

            fec_rt = mfem.RT_FECollection(order-1, dim)
            fes_rt = mfem.FiniteElementSpace(mesh, fec_rt)

            dsol = mfem.GridFunction(fes_rt)

            dsol.ProjectCoefficient(dsolCoef)

            sol_sock.precision(8)
            dsol_sock.precision(8)
            sol_sock << "solution\n" << mesh << sol << "window_title 'Solution'"
            sol_sock.flush()
            dsol_sock << "solution\n" << mesh << dsol
            dsol_sock.flush()
            dsol_sock << "window_geometry 0 375 400 350 " << "window_title 'Curl of solution'"
            dsol_sock.flush()


def E_exact(x):
    E = np.zeros(shape, dtype=np.float64)
    if dim == 1:
        E[0] = 1.1 * sin(kappa * x[0] + 0.0 * pi)
        E[1] = 1.2 * sin(kappa * x[0] + 0.4 * pi)
        E[2] = 1.3 * sin(kappa * x[0] + 0.9 * pi)
    elif dim == 2:
        E[0] = 1.1 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        E[1] = 1.2 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        E[2] = 1.3 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
    else:
        E[0] = 1.1 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        E[1] = 1.2 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        E[2] = 1.3 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)

        for i in range(3):
            E[i] = E[i]*cos(kappa * x[2])
    return E


def CurlE_exact(x):
    dE = np.zeros(shape, dtype=np.float64)
    if dim == 1:
        c4 = cos(kappa * x[0] + 0.4 * pi)
        c9 = cos(kappa * x[0] + 0.9 * pi)

        dE[0] = 0.0
        dE[1] = -1.3 * c9
        dE[2] = 1.2 * c4
        for i in range(3):
            dE[i] = dE[i]*kappa

    elif dim == 2:
        c0 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        c4 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c9 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)

        dE[0] = 1.3 * c9
        dE[1] = -1.3 * c9
        dE[2] = 1.2 * c4 - 1.1 * c0
        for i in range(3):
            dE[i] = dE[i]*kappa*sqrt1_2

    else:
        s0 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        c0 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        s4 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c4 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c9 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
        sk = sin(kappa * x[2])
        ck = cos(kappa * x[2])

        dE[0] = 1.2 * s4 * sk + 1.3 * sqrt1_2 * c9 * ck
        dE[1] = -1.1 * s0 * sk - 1.3 * sqrt1_2 * c9 * ck
        dE[2] = -sqrt1_2 * (1.1 * c0 - 1.2 * c4) * ck
        for i in range(3):
            dE[i] = dE[i]*kappa
    return dE


def f_exact(x):
    f = np.zeros(shape, dtype=np.float64)
    if dim == 1:
        s0 = sin(kappa * x[0] + 0.0 * pi)
        s4 = sin(kappa * x[0] + 0.4 * pi)
        s9 = sin(kappa * x[0] + 0.9 * pi)

        f[0] = 2.2 * s0 + 1.2 * sqrt1_2 * s4
        f[1] = (1.2 * (2.0 + kappa * kappa) * s4 +
                sqrt1_2 * (1.1 * s0 + 1.3 * s9))
        f[2] = 1.3 * (2.0 + kappa * kappa) * s9 + 1.2 * sqrt1_2 * s4

    elif dim == 2:
        s0 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        s4 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        s9 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)

        f[0] = (0.55 * (4.0 + kappa * kappa) * s0 +
                0.6 * (sqrt2 - kappa * kappa) * s4)
        f[1] = (0.55 * (sqrt2 - kappa * kappa) * s0 +
                0.6 * (4.0 + kappa * kappa) * s4 +
                0.65 * sqrt2 * s9)
        f[2] = 0.6 * sqrt2 * s4 + 1.3 * (2.0 + kappa * kappa) * s9

    else:
        s0 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        c0 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        s4 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c4 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        s9 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
        c9 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
        sk = sin(kappa * x[2])
        ck = cos(kappa * x[2])

        f[0] = (0.55 * (4.0 + 3.0 * kappa * kappa) * s0 * ck +
                0.6 * (sqrt2 - kappa * kappa) * s4 * ck -
                0.65 * sqrt2 * kappa * kappa * c9 * sk)

        f[1] = (0.55 * (sqrt2 - kappa * kappa) * s0 * ck +
                0.6 * (4.0 + 3.0 * kappa * kappa) * s4 * ck +
                0.65 * sqrt2 * s9 * ck -
                0.65 * sqrt2 * kappa * kappa * c9 * sk)

        f[2] = (0.6 * sqrt2 * s4 * ck -
                sqrt2 * kappa * kappa * (0.55 * c0 + 0.6 * c4) * sk
                + 1.3 * (2.0 + kappa * kappa) * s9 * ck)
    return f


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex31 (Definite Maxwell Problem)')
    parser.add_argument('-m', '--mesh',
                        default="inline-quad.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-r', '--refine',
                        action='store', default=2, type=int,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=1.0,
                        help="Set the frequency for the exact")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        assert False, "This example requires numba to run"
    parser.add_argument("-n", "--numba",
                        default=int(HAS_NUMBA),
                        type=int,
                        help="Use Number compiled coefficient")

    args = parser.parse_args()
    args.numba = bool(args.numba)
    parser.print_options(args)

    order = args.order

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    freq = args.frequency
    numba = args.numba

    run(freq=freq,
        order=order,
        refine=args.refine,
        meshfile=meshfile,
        visualization=visualization,
        numba=numba)



'''
   PyMFEM example 37

   See c++ version in the MFEM library for more detail

   Sample runs:
       python ex37.py -alpha 10
       python ex37.py -alpha 10 -pv
       python ex37.py -lambda 0.1 -mu 0.1
       python ex37.py -o 2 -alpha 5.0 -mi 50 -vf 0.4 -ntol 1e-5
       python ex37.py -r 6 -o 1 -alpha 25.0 -epsilon 0.02 -mi 50 -ntol 1e-5


   Description: This example code demonstrates the use of MFEM to solve a
                density-filtered [3] topology optimization problem. The
                objective is to minimize the compliance

                    minimize ∫_Ω f⋅u dx over u ∈ [H¹(Ω)]² and ρ ∈ L¹(Ω)

                    subject to

                      -Div(r(ρ̃)Cε(u)) = f       in Ω + BCs
                      -ϵ²Δρ̃ + ρ̃ = ρ             in Ω + Neumann BCs
                      0 ≤ ρ ≤ 1                 in Ω
                      ∫_Ω ρ dx = θ vol(Ω)

                Here, r(ρ̃) = ρ₀ + ρ̃³ (1-ρ₀) is the solid isotropic material
                penalization (SIMP) law, C is the elasticity tensor for an
                isotropic linearly elastic material, ϵ > 0 is the design
                length scale, and 0 < θ < 1 is the volume fraction.

                The problem is discretized and gradients are computing using
                finite elements [1]. The design is optimized using an entropic
                mirror descent algorithm introduced by Keith and Surowiec [2]
                that is tailored to the bound constraint 0 ≤ ρ ≤ 1.

                This example highlights the ability of MFEM to deliver high-
                order solutions to inverse design problems and showcases how
                to set up and solve PDE-constrained optimization problems
                using the so-called reduced space approach.

   [1] Andreassen, E., Clausen, A., Schevenels, M., Lazarov, B. S., & Sigmund, O.
       (2011). Efficient topology optimization in MATLAB using 88 lines of
       code. Structural and Multidisciplinary Optimization, 43(1), 1-16.
   [2] Keith, B. and Surowiec, T. (2023) Proximal Galerkin: A structure-
       preserving finite element method for pointwise bound constraints.
       arXiv:2307.12444 [math.NA]
   [3] Lazarov, B. S., & Sigmund, O. (2011). Filters in topology optimization
       based on Helmholtz‐type differential equations. International Journal
       for Numerical Methods in Engineering, 86(6), 765-781.

'''
import mfem.par as mfem
from mfem.par import intArray, doubleArray

from ex37_common import (sigmoid,
                         der_sigmoid,
                         inv_sigmoid,
                         MappedGridFunctionCoefficient,
                         DiffMappedGridFunctionCoefficient,
                         SIMPInterpolationCoefficient,
                         VolumeForceCoefficient,
                         )
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt, floor
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)

visualization = True
paraview_output = False


class StrainEnergyDensityCoefficient():
    '''
    Python Note: 
       Approach here is to replace Eval and GetVectorGradient method call in C++
       using the dependency feature of mfem.jit.

       GetVectorGradient is mimiced by creating GradientGridFunctionCoefficient
       for each component of u vector. Note GridFunction(fes, u.GetDataArray())
       reuses the data array from u.
    '''

    def __init__(self, llambda, mu, u, rho_filter, rho_min=1e-6, exponent=3.0):
        assert rho_min >= 0.0, "rho_min must be >= 0"
        assert rho_min < 1.0,  "rho_min must be > 1"

        pfes = u.ParFESpace()
        assert pfes.GetOrdering() == mfem.Ordering.byNODES, "u has to use byNODES ordering"

        pmesh = pfes.GetParMesh()
        dim = pmesh.Dimension()
        assert dim == 2, "dim must be two."

        fec = pfes.FEColl()
        pfes = mfem.ParFiniteElementSpace(pmesh, fec)
        size = len(u.GetDataArray())

        u1 = mfem.ParGridFunction(pfes, mfem.Vector(
            u.GetDataArray()))   # first component
        u2 = mfem.ParGridFunction(pfes, mfem.Vector(
            u.GetDataArray()), size//2)  # second component

        c_gradu1 = mfem.GradientGridFunctionCoefficient(u1)
        c_gradu2 = mfem.GradientGridFunctionCoefficient(u2)

        c_rho_filter = mfem.GridFunctionCoefficient(rho_filter)

        @mfem.jit.scalar(dependency=(llambda, mu, c_gradu1, c_gradu2, c_rho_filter))
        def coeff(ptx, L, M, grad1, grad2, val):
            div_u = grad1[0] + grad2[1]
            density = L*div_u*div_u

            grad = np.zeros(shape=(2, 2), dtype=np.float64)
            grad[0, 0] = grad1[0]
            grad[0, 1] = grad1[1]
            grad[1, 0] = grad2[0]
            grad[1, 1] = grad2[1]

            for i in range(2):
                for j in range(2):
                    density += M*grad[i, j]*(grad[i, j] + grad[j, i])
            return -exponent * val**(exponent-1.0)*(1-rho_min)*density

        self.pfes = pfes
        self.size = size
        self.u1u2 = (u1, u2)
        self.dependency = (c_gradu1, c_gradu2, c_rho_filter)
        self.coeff = coeff

    def Update(self, u, rho_filter):
        u1 = mfem.ParGridFunction(self.pfes, mfem.Vector(u.GetDataArray()))
        u2 = mfem.ParGridFunction(self.pfes, mfem.Vector(u.GetDataArray()),
                                  self.size//2)
        self.dependency[0].SetGridFunction(u1)
        self.dependency[1].SetGridFunction(u2)
        self.dependency[2].SetGridFunction(rho_filter)
        self.u1u2 = (u1, u2)


class DiffusionSolver():
    def __init__(self):
        self.rhscf = None
        self.neumann_cf = None
        self.masscf = None
        self.essbdr_cf = None
        self.gradient_cf = None

    def SetupFEM(self):
        dim = self.pmesh.Dimension()
        self.fec = mfem.H1_FECollection(self.order, dim)
        self.pfes = mfem.ParFiniteElementSpace(self.pmesh, self.fec)

        if self.ess_bdr.Size() == 0 and self.pmesh.bdr_attributes.Size() > 0:
            self.ess_bdr = mfem.intArray([1]*self.pmesh.bdr_attributes.Max())

    def Solve(self):
        A = mfem.OperatorPtr()
        B = mfem.Vector()
        X = mfem.Vector()
        ess_tdof_list = mfem.intArray()
        self.pfes.GetEssentialTrueDofs(self.ess_bdr, ess_tdof_list)

        self.u = mfem.ParGridFunction(self.pfes)
        self.u.Assign(0.0)

        b = mfem.ParLinearForm(self.pfes)

        if self.rhscf is not None:
            itg = mfem.DomainLFIntegrator(self.rhscf)
            b.AddDomainIntegrator(itg)

        if self.neumann_cf is not None:
            assert self.neumann_bdr.Size() > 0, "neumann_bdr attributes not provided"
            b.AddBoundaryIntegrator(mfem.BoundaryLFIntegrator(self.neumann_cf),
                                    self.neumann_bdr)
        elif self.gradient_cf is not None:
            assert self.neumann_bdr.Size() > 0, "neumann_bdr attributes not provided"
            b.AddBoundaryIntegrator(mfem.BoundaryNormalLFIntegrator(self.gradient_cf),
                                    self.neumann_bdr)

        b.Assemble()

        a = mfem.ParBilinearForm(self.pfes)
        a.AddDomainIntegrator(mfem.DiffusionIntegrator(self.diffcf))
        if self.masscf is not None:
            a.AddDomainIntegrator(mfem.MassIntegrator(self.masscf))
        a.Assemble()
        if self.essbdr_cf is not None:
            self.u.ProjectBdrCoefficient(essbdr_cf, ess_bdr)

        a.FormLinearSystem(ess_tdof_list, self.u, b, A, X, B)

        M = mfem.HypreBoomerAMG()
        M.SetPrintLevel(0)
        cg = mfem.CGSolver(MPI.COMM_WORLD)
        cg.SetRelTol(1e-12)
        cg.SetMaxIter(10000)
        cg.SetPrintLevel(0)
        cg.SetPreconditioner(M)
        cg.SetOperator(A.Ptr())
        cg.Mult(B, X)

        a.RecoverFEMSolution(X, b, self.u)
        self.b = b

    def GetFEMSolution(self):
        return self.u

    def SetRHSCoefficient(self, rhscf):
        self.rhscf = rhscf


class LinearElasticitySolver():
    def __init__(self):
        self.rhs_cf = None
        self.essbdr_cf = None

    def SetupFEM(self):
        dim = self.pmesh.Dimension()
        self.fec = mfem.H1_FECollection(
            self.order, dim, mfem.BasisType.Positive)
        self.pfes = mfem.ParFiniteElementSpace(self.pmesh, self.fec, dim)
        self.u = mfem.ParGridFunction(self.pfes)
        self.u.Assign(0.0)

    def Solve(self):
        A = mfem.OperatorPtr()
        B = mfem.Vector()
        X = mfem.Vector()

        ess_tdof_list = mfem.intArray()
        self.pfes.GetEssentialTrueDofs(self.ess_bdr, ess_tdof_list)

        x = mfem.ParGridFunction(self.pfes)
        x .Assign(0.0)

        self.u.Assign(0.0)
        b = mfem.ParLinearForm(self.pfes)

        if self.rhs_cf is not None:
            b.AddDomainIntegrator(mfem.VectorDomainLFIntegrator(self.rhs_cf))
        b.Assemble()

        a = mfem.ParBilinearForm(self.pfes)
        a.AddDomainIntegrator(
            mfem.ElasticityIntegrator(self.lambda_cf, self.mu_cf))
        a.Assemble()
        if self.essbdr_cf is not None:
            u.ProjectBdrCoefficient(self.essbdr_cf, self.ess_bdr)

        a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

        M = mfem.HypreBoomerAMG()
        M.SetPrintLevel(0)
        cg = mfem.CGSolver(MPI.COMM_WORLD)

        cg.SetRelTol(1e-10)
        cg.SetMaxIter(10000)
        cg.SetPrintLevel(0)
        cg.SetPreconditioner(M)
        cg.SetOperator(A.Ptr())
        cg.Mult(B, X)
        a.RecoverFEMSolution(X, b, x)

        self.u += x
        self.b = b

    def GetFEMSolution(self):
        return self.u

    def GetParLinearForm(self):
        return self.b


class Proj():
    '''

    @brief Bregman projection of ρ = sigmoid(ψ) onto the subspace
          ∫_Ω ρ dx = θ vol(Ω) as follows:

          1. Compute the root of the R → R function
              f(c) = ∫_Ω sigmoid(ψ + c) dx - θ vol(Ω)
          2. Set ψ ← ψ + c.

    @param psi a GridFunction to be updated
    @param target_volume θ vol(Ω)
    @param tol Newton iteration tolerance
    @param max_its Newton maximum iteration number
    @return double Final volume, ∫_Ω sigmoid(ψ)

    '''

    def __init__(self, psi):
        self.psi = psi
        self.sigmoid_psi = MappedGridFunctionCoefficient(psi, sigmoid)
        self.der_sigmoid_psi = MappedGridFunctionCoefficient(psi, der_sigmoid)

    def __call__(self, target_volume, tol=1e-12, max_its=10):
        psi = self.psi
        int_sigmoid_psi = mfem.ParLinearForm(psi.ParFESpace())
        int_sigmoid_psi.AddDomainIntegrator(
            mfem.DomainLFIntegrator(self.sigmoid_psi))
        int_der_sigmoid_psi = mfem.ParLinearForm(psi.ParFESpace())
        int_der_sigmoid_psi.AddDomainIntegrator(mfem.DomainLFIntegrator(
            self.der_sigmoid_psi))
        done = False
        for k in range(max_its):  # Newton iteration
            int_sigmoid_psi.Assemble()         # Recompute f(c) with updated ψ
            f = int_sigmoid_psi.Sum()
            f = MPI.COMM_WORLD.allreduce(f, MPI.SUM)
            f -= target_volume

            int_der_sigmoid_psi.Assemble()      # Recompute df(c) with updated ψ
            df = int_der_sigmoid_psi.Sum()
            df = MPI.COMM_WORLD.allreduce(df, MPI.SUM)
            dc = -f/df
            psi += dc
            if abs(dc) < tol:
                done = True
                break

        if not done:
            message = ("Projection reached maximum iteration without converging. " +
                       "Result may not be accurate.")
            import warnigns
            warnings.warn(message, RuntimeWarning)

        int_sigmoid_psi.Assemble()
        material_volume = int_sigmoid_psi.Sum()
        material_volume = MPI.COMM_WORLD.allreduce(material_volume, MPI.SUM)
        return material_volume


def run(ref_levels=5,
        order=2,
        alpha=1.0,
        epsilon=0.01,
        vol_fraction=0.5,
        max_it=1e3,
        itol=1e-1,
        ntol=1e-4,
        rho_min=1e-6,
        llambda=1.0,
        mu=1.0,):

    mesh = mfem.Mesh.MakeCartesian2D(3, 1, mfem.Element.QUADRILATERAL,
                                     True, 3.0, 1.0)
    dim = mesh.Dimension()

    # 2. Set BCs.
    for i in range(mesh.GetNBE()):
        be = mesh.GetBdrElement(i)
        vertices = mesh.GetBdrElementVertices(i)   # this method returns list

        coords1 = mesh.GetVertexArray(vertices[0])   # this returns numpy array
        coords2 = mesh.GetVertexArray(vertices[1])

        center = (coords1 + coords2)/2

        if abs(center[0] - 0.0) < 1e-10:
            # the left edge
            be.SetAttribute(1)
        else:
            # all other boundaries
            be.SetAttribute(2)
    mesh.SetAttributes()

    # 3. Refine the mesh.
    for lev in range(ref_levels):
        mesh.UniformRefinement()
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    mesh.Clear()

    # 4. Define the necessary finite element spaces on the mesh.
    state_fec = mfem.H1_FECollection(order, dim)    # space for u
    filter_fec = mfem.H1_FECollection(order, dim)   # space for ρ̃
    control_fec = mfem.L2_FECollection(order-1, dim,
                                       mfem.BasisType.GaussLobatto)  # space for ψ
    state_fes = mfem.ParFiniteElementSpace(pmesh, state_fec, dim)
    filter_fes = mfem.ParFiniteElementSpace(pmesh, filter_fec)
    control_fes = mfem.ParFiniteElementSpace(pmesh, control_fec)

    state_size = state_fes.GlobalTrueVSize()
    control_size = control_fes.GlobalTrueVSize()
    filter_size = filter_fes.GlobalTrueVSize()

    if myid == 0:
        print("Number of state unknowns: " + str(state_size))
        print("Number of filter unknowns: " + str(filter_size))
        print("Number of control unknowns: " + str(control_size))

    #  5. Set the initial guess for ρ.
    u = mfem.ParGridFunction(state_fes)
    psi = mfem.ParGridFunction(control_fes)
    psi_old = mfem.ParGridFunction(control_fes)
    rho_filter = mfem.ParGridFunction(filter_fes)

    u.Assign(0.0)
    rho_filter.Assign(vol_fraction)
    psi.Assign(inv_sigmoid(vol_fraction))
    psi_old.Assign(inv_sigmoid(vol_fraction))

    # ρ = sigmoid(ψ)
    rho = MappedGridFunctionCoefficient(psi, sigmoid)
    #  Interpolation of ρ = sigmoid(ψ) in control fes (for ParaView output)
    rho_gf = mfem.ParGridFunction(control_fes)
    # ρ - ρ_old = sigmoid(ψ) - sigmoid(ψ_old)
    succ_diff_rho = DiffMappedGridFunctionCoefficient(psi, psi_old, sigmoid)

    # 6. Set-up the physics solver.
    maxat = pmesh.bdr_attributes.Max()
    ess_bdr = intArray([0]*maxat)
    ess_bdr[0] = 1

    one = mfem.ConstantCoefficient(1.0)
    lambda_cf = mfem.ConstantCoefficient(llambda)
    mu_cf = mfem.ConstantCoefficient(mu)

    ElasticitySolver = LinearElasticitySolver()
    ElasticitySolver.pmesh = pmesh
    ElasticitySolver.order = state_fec.GetOrder()
    ElasticitySolver.SetupFEM()

    center = np.array((2.9, 0.5))
    force = np.array((0.0, -1.0))
    r = 0.05
    vforce_cf = VolumeForceCoefficient(r, center, force)
    ElasticitySolver.rhs_cf = vforce_cf
    ElasticitySolver.ess_bdr = ess_bdr

    # 7. Set-up the filter solver.
    eps2_cf = mfem.ConstantCoefficient(epsilon*epsilon)
    FilterSolver = DiffusionSolver()
    FilterSolver.pmesh = pmesh
    FilterSolver.order = filter_fec.GetOrder()
    FilterSolver.diffcf = eps2_cf
    FilterSolver.masscf = one

    if pmesh.bdr_attributes.Size() > 0:
        ess_bdr_filter = mfem.intArray([0]*pmesh.bdr_attributes.Max())
    else:
        ess_bdr_filter = mfem.intArray()

    FilterSolver.ess_bdr = ess_bdr_filter
    FilterSolver.SetupFEM()

    mass = mfem.ParBilinearForm(control_fes)
    mass.AddDomainIntegrator(mfem.InverseIntegrator(mfem.MassIntegrator(one)))
    mass.Assemble()
    M = mfem.HypreParMatrix()
    empty = mfem.intArray()
    mass.FormSystemMatrix(empty, M)

    # 8. Define the Lagrange multiplier and gradient functions.
    grad = mfem.ParGridFunction(control_fes)
    w_filter = mfem.ParGridFunction(filter_fes)

    # 9. Define some tools for later.
    zero = mfem.ConstantCoefficient(0.0)
    onegf = mfem.ParGridFunction(control_fes)
    onegf.Assign(1.0)
    zerogf = mfem.ParGridFunction(control_fes)
    zerogf.Assign(0.0)

    vol_form = mfem.ParLinearForm(control_fes)
    vol_form.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
    vol_form.Assemble()
    domain_volume = vol_form(onegf)
    target_volume = domain_volume * vol_fraction

    # 10. Connect to GLVis. Prepare for VisIt output.
    if visualization:
        sout_r = mfem.socketstream("localhost", 19916)
        sout_r.precision(8)

    if paraview_output:
        paraview_dc = mfem.ParaViewDataCollection("ex37p", pmesh)

        rho_gf.ProjectCoefficient(rho)
        paraview_dc.SetPrefixPath("ParaView")
        paraview_dc.SetLevelsOfDetail(order)
        paraview_dc.SetDataFormat(mfem.VTKFormat.BINARY)
        paraview_dc.SetHighOrderOutput(True)
        paraview_dc.SetCycle(0)
        paraview_dc.SetTime(0.0)
        paraview_dc.RegisterField("displacement", u)
        paraview_dc.RegisterField("density", rho_gf)
        paraview_dc.RegisterField("filtered_density", rho_filter)
        paraview_dc.Save()

    # 11. Iterate:
    proj = Proj(psi)
    for k in range(1, max_it+1):
        if k > 1:
            alpha *= k/(k-1.)

        if myid == 0:
            print("\nStep = " + str(k))

        # Step 1 - Filter solve
        # Solve (ϵ^2 ∇ ρ̃, ∇ v ) + (ρ̃,v) = (ρ,v)
        FilterSolver.SetRHSCoefficient(rho)
        FilterSolver.Solve()
        rho_filter = FilterSolver.GetFEMSolution()

        # Step 2 - State solve
        # Solve (λ r(ρ̃) ∇⋅u, ∇⋅v) + (2 μ r(ρ̃) ε(u), ε(v)) = (f,v)
        if k == 1:
            SIMP = SIMPInterpolationCoefficient(rho_filter, rho_min, 1.0)
            SIMP_cf = SIMP.coeff
        else:
            SIMP.Update(rho_filter)

        lambda_SIMP_cf = mfem.ProductCoefficient(lambda_cf, SIMP_cf)
        mu_SIMP_cf = mfem.ProductCoefficient(mu_cf, SIMP_cf)

        ElasticitySolver.lambda_cf = lambda_SIMP_cf
        ElasticitySolver.mu_cf = mu_SIMP_cf

        ElasticitySolver.Solve()
        u = ElasticitySolver.GetFEMSolution()

        # Step 3 - Adjoint filter solve
        # Solve (ϵ² ∇ w̃, ∇ v) + (w̃ ,v) = (-r'(ρ̃) ( λ |∇⋅u|² + 2 μ |ε(u)|²),v)
        if k == 1:
            SEDC = StrainEnergyDensityCoefficient(lambda_cf, mu_cf, u, rho_filter,
                                                  rho_min)
            rhs_cf = SEDC.coeff
        else:
            SEDC.Update(u, rho_filter)

        FilterSolver.SetRHSCoefficient(rhs_cf)
        FilterSolver.Solve()
        w_filter = FilterSolver.GetFEMSolution()

        # Step 4 - Compute gradient
        # Solve G = M⁻¹w̃
        w_cf = mfem.GridFunctionCoefficient(w_filter)
        w_rhs = mfem.ParLinearForm(control_fes)
        w_rhs.AddDomainIntegrator(mfem.DomainLFIntegrator(w_cf))
        w_rhs.Assemble()
        M.Mult(w_rhs, grad)

        # Step 5 - Update design variable ψ ← proj(ψ - αG)
        psi.Add(-alpha, grad)
        material_volume = proj(target_volume)

        # Compute ||ρ - ρ_old|| in control fes.
        norm_increment = zerogf.ComputeL1Error(succ_diff_rho)
        norm_reduced_gradient = norm_increment/alpha
        psi_old.Assign(psi)

        compliance = (ElasticitySolver.GetParLinearForm())(u)
        #compliance = MPI.COMM_WORLD.allreduce(compliance, MPI.SUM)

        if myid == 0:
            print("norm of the reduced gradient = " +
                  "{:g}".format(norm_reduced_gradient))
            print("norm of the increment = " + "{:g}".format(norm_increment))
            print("compliance = " + "{:g}".format(compliance))
            print("volume fraction = " +
                  "{:g}".format(material_volume / domain_volume))

        if visualization:
            r_gf = mfem.ParGridFunction(filter_fes)
            r_gf.ProjectCoefficient(SIMP_cf)
            sout_r << "parallel " << num_procs << " " << myid << "\n"
            sout_r << "solution\n" << pmesh << r_gf << "window_title 'Design density r(ρ̃)'"
            sout_r.flush()

        if paraview_output:
            rho_gf.ProjectCoefficient(rho)
            paraview_dc.SetCycle(k)
            paraview_dc.SetTime(float(k))
            paraview_dc.Save()

        if norm_reduced_gradient < ntol and norm_increment < itol:
            break


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex37p (Topology Optimization)')

    parser.add_argument('-r', '--refine',
                        action='store', default=5, type=int,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument("-o", "--order",
                        action='store', default=2, type=int,
                        help="Order (degree) of the finite elements.")
    parser.add_argument("-alpha", "--alpha-step-length",
                        action='store', default=1.0, type=float,
                        help="Step length for gradient descent.")
    parser.add_argument("-epsilon", "--epsilon-thickness",
                        action='store', default=0.01, type=float,
                        help="Length scale for ρ.")
    parser.add_argument("-mi", "--max-it",
                        action='store', default=1000, type=int,
                        help="Maximum number of gradient descent iterations.")
    parser.add_argument("-ntol", "--rel-tol",
                        action='store', default=1e-4, type=float,
                        help="Normalized exit tolerance.")
    parser.add_argument("-itol", "--abs-tol",
                        action='store', default=1e-1, type=float,
                        help="Increment exit tolerance.")
    parser.add_argument("-vf", "--volume-fraction",
                        action='store', default=0.5, type=float,
                        help="Volume fraction for the material density.")
    parser.add_argument("-lambda", "--llambda",
                        action='store', default=1.0, type=float,
                        help="Lamé constant λ.")
    parser.add_argument("-mu", "--mu",
                        action='store', default=1.0, type=float,
                        help="Lamé constant μ.")
    parser.add_argument("-rmin", "--psi-min",
                        action='store', default=1e-6, type=float,
                        help="Minimum of density coefficient.")
    parser.add_argument("-pv", "--paraview",
                        action='store_true', default=False,
                        help="Enable or disable ParaView output.")
    parser.add_argument("-no-vis", "--no-visualization",
                        action='store_true', default=False,
                        help='Enable GLVis visualization')

    args = parser.parse_args()

    if myid == 0:
        parser.print_options(args)

    globals()["visualization"] = not args.no_visualization
    globals()["paraview_output"] = args.paraview

    run(ref_levels=args.refine,
        order=args.order,
        alpha=args.alpha_step_length,
        epsilon=args.epsilon_thickness,
        vol_fraction=args.volume_fraction,
        max_it=args.max_it,
        itol=args.abs_tol,
        ntol=args.rel_tol,
        rho_min=args.psi_min,
        llambda=args.llambda,
        mu=args.mu)



from setuptools import build_meta as _orig
from setuptools.build_meta import *

import build_globals as bglb


def get_requires_for_build_wheel(config_settings=None):
    ret = _orig.get_requires_for_build_wheel(config_settings)

    need_mpi = False
    if config_settings is not None:
        for flag in ("with-parallel", ):
            value = config_settings.pop(flag, "No")
            if value.upper() in ("YES", "TRUE", "1"):
                need_mpi = True
                break

    if need_mpi:
        ret = ret + ['mpi4py']
    return ret


def get_requires_for_build_sdist(config_settings=None):
    return _orig.get_requires_for_build_sdist(config_settings)


def build_wheel(*args, **kwargs):
    bglb.cfs = args[1]
    if bglb.cfs is None:
        bglb.cfs = {}
    return _orig.build_wheel(*args, **kwargs)



file = "../_ser/lininteg.py"

out = ["namespace mfem {"]
fid = open(file, 'r')
for line in fid.readlines():
    if line.startswith("class"):
        cname = (line.split(' ')[1]).split('(')[0]

    if line.startswith("    def __init__"):
        pp = []
        if line.find(", ir=None") != -1:
            line = line.replace(", ir=None", "")
            pp.append("    self._ir=ir")
        if line.find("*args") != -1:
            pp.append("    self._coeff = args")
        elif line.find("self, seed_=0") != -1:
            pp.append("    self._coeff = QG")
        elif line.find("self, vqfc") != -1:
            pp.append("    self._coeff = vqfc")
        elif line.find("self, qfc") != -1:
            pp.append("    self._coeff = qfc")
        elif line.find(", QG") != -1:
            pp.append("    self._coeff = QG")
        elif line.find(", QF)") != -1:
            pp.append("    self._coeff = QF")
        elif line.find(", F)") != -1:
            pp.append("    self._coeff = F")
        elif line.find(", f)") != -1:
            pp.append("    self._coeff = f")
        elif line.find(", f, s=1.0)") != -1:
            pp.append("    self._coeff = f")
        elif line.find(", uD_, lambda_, mu_, alpha_, kappa_)") != -1:
            pp.append("    self._coeff = uD_")
        elif line.find("(self)") != -1:
            pass
        else:
            print(cname)
            print(line)
            assert False, "No recipt for this pattern "
        if len(pp) > 0:
            out.append("%pythonappend " + cname + "::" + cname + " %{")
            for x in pp:
                out.append(x)
            out.append("%}")
fid.close()
out.append("}")
fid = open("lininteg_ext.i", "w")
fid.write("\n".join(out))
fid.close()




# ----------------------------------------------------------------------------------------
# Routines for PyMFEM Wrapper Generation/Compile
# ----------------------------------------------------------------------------------------
import sys
import os
import re
import subprocess

__all__ = ["cmake_make_mfem"]

from build_consts import *
from build_utils import *

import build_globals as bglb


def cmake_make_mfem(serial=True):
    '''
    build MFEM
    '''
    cmbuild = 'cmbuild_ser' if serial else 'cmbuild_par'
    path = os.path.join(extdir, 'mfem', cmbuild)
    if os.path.exists(path):
        print("working directory already exists!")
    else:
        os.makedirs(path)

    ldflags = os.getenv('LDFLAGS') if os.getenv('LDFLAGS') is not None else ''
    metisflags = ''
    hypreflags = ''

    rpaths = []
    if sys.platform in ("linux", "linux2"):
        rpaths_origin = "$ORIGIN"
    elif sys.platform == "darwin":
        rpaths_origin = "@loader_path"

    def add_rpath(p, dest):
        p = os.path.join(rpaths_origin, os.path.relpath(p, dest))
        
        if not p in rpaths:
            rpaths.append(p)

    cmake_opts = {'DBUILD_SHARED_LIBS': '1',
                  'DMFEM_ENABLE_EXAMPLES': '1',
                  'DMFEM_ENABLE_MINIAPPS': '0',
                  'DCMAKE_SHARED_LINKER_FLAGS': ldflags,
                  'DMFEM_USE_ZLIB': '1',
                  'DCMAKE_CXX_FLAGS': bglb.cxxstd_flag,
                  'DCMAKE_BUILD_WITH_INSTALL_RPATH': '1'}

    if sys.platform == 'darwin':
        cmake_opts["DCMAKE_MACOSX_RPATH"] = 'YES'
        cmake_opts["DCMAKE_INSTALL_NAME_DIR"] = '@rpath'        

    if bglb.mfem_debug:
        cmake_opts['DMFEM_DEBUG'] = 'YES'

    if bglb.mfem_build_miniapps:
        cmake_opts['DMFEM_ENABLE_MINIAPPS'] = '1'

    if bglb.verbose:
        cmake_opts['DCMAKE_VERBOSE_MAKEFILE'] = '1'

    if serial:
        ex_loc = os.path.join(bglb.mfems_prefix, "examples")        
        cmake_opts['DCMAKE_CXX_COMPILER'] = bglb.cxx_command
        cmake_opts['DMFEM_USE_EXCEPTIONS'] = '1'
        cmake_opts['DCMAKE_INSTALL_PREFIX'] = bglb.mfems_prefix

        add_rpath(os.path.join(bglb.mfems_prefix, 'lib'), ex_loc)
        if bglb.enable_suitesparse:
            enable_metis = True
        else:
            enable_metis = False
        #assert False, rpaths            
    else:
        ex_loc = os.path.join(bglb.mfemp_prefix, "examples")
        cmake_opts['DCMAKE_CXX_COMPILER'] = bglb.mpicxx_command
        cmake_opts['DMFEM_USE_EXCEPTIONS'] = '0'
        cmake_opts['DCMAKE_INSTALL_PREFIX'] = bglb.mfemp_prefix
        cmake_opts['DMFEM_USE_MPI'] = '1'
        cmake_opts['DHYPRE_DIR'] = bglb.hypre_prefix
        cmake_opts['DHYPRE_INCLUDE_DIRS'] = os.path.join(
            bglb.hypre_prefix, "include")

        add_rpath(os.path.join(bglb.mfemp_prefix, 'lib'), ex_loc)

        hyprelibpath = os.path.dirname(
            find_libpath_from_prefix(
                'HYPRE', bglb.hypre_prefix))

        add_rpath(hyprelibpath, ex_loc)

        hypreflags = "-L" + hyprelibpath + " -lHYPRE "

        if bglb.enable_strumpack:
            cmake_opts['DMFEM_USE_STRUMPACK'] = '1'
            cmake_opts['DSTRUMPACK_DIR'] = bglb.strumpack_prefix
            libpath = os.path.dirname(
                find_libpath_from_prefix("STRUMPACK", bglb.strumpack_prefix))
            add_rpath(libpath, ex_loc)
        if bglb.enable_pumi:
            cmake_opts['DMFEM_USE_PUMI'] = '1'
            cmake_opts['DPUMI_DIR'] = bglb.pumi_prefix
            libpath = os.path.dirname(
                find_libpath_from_prefix("pumi", bglb.strumpack_prefix))
            add_rpath(libpath, ex_loc)
        enable_metis = True

    if enable_metis:
        cmake_opts['DMFEM_USE_METIS_5'] = '1'
        cmake_opts['DMETIS_DIR'] = bglb.metis_prefix
        cmake_opts['DMETIS_INCLUDE_DIRS'] = os.path.join(
            bglb.metis_prefix, "include")
        metislibpath = os.path.dirname(
            find_libpath_from_prefix(
                'metis', bglb.metis_prefix))
        add_rpath(metislibpath, ex_loc)

        if bglb.use_metis_gklib:
            metisflags = "-L" + metislibpath + " -lmetis -lGKlib "
        else:
            metisflags = "-L" + metislibpath + " -lmetis "

    if ldflags != '':
        cmake_opts['DCMAKE_SHARED_LINKER_FLAGS'] = ldflags
        cmake_opts['DCMAKE_EXE_LINKER_FLAGS'] = ldflags

    if metisflags != '':
        cmake_opts['DMETIS_LIBRARIES'] = metisflags
    if hypreflags != '':
        cmake_opts['DHYPRE_LIBRARIES'] = hypreflags

    if bglb.enable_cuda:
        cmake_opts['DMFEM_USE_CUDA'] = '1'
        if bglb.cuda_arch != '':
            cmake_opts['DCMAKE_CUDA_ARCHITECTURES'] = bglb.cuda_arch

    if bglb.enable_libceed:
        cmake_opts['DMFEM_USE_CEED'] = '1'
        cmake_opts['DCEED_DIR'] = bglb.libceed_prefix
        libpath = os.path.dirname(
            find_libpath_from_prefix("ceed", bglb.libceed_prefix))
        add_rpath(libpath, ex_loc)

    if bglb.enable_gslib:
        if serial:
            cmake_opts['DMFEM_USE_GSLIB'] = '1'
            cmake_opts['DGSLIB_DIR'] = bglb.gslibs_prefix
        else:
            cmake_opts['DMFEM_USE_GSLIB'] = '1'
            cmake_opts['DGSLIB_DIR'] = bglb.gslibp_prefix

    if bglb.enable_suitesparse:
        cmake_opts['DMFEM_USE_SUITESPARSE'] = '1'
        if bglb.suitesparse_prefix != '':
            cmake_opts['DSuiteSparse_DIR'] = bglb.suitesparse_prefix

    if bglb.enable_lapack:
        cmake_opts['DMFEM_USE_LAPACK'] = '1'
    if bglb.blas_libraries != "":
        cmake_opts['DBLAS_LIBRARIES'] = bglb.blas_libraries
    if bglb.lapack_libraries != "":
        cmake_opts['DLAPACK_LIBRARIES'] = bglb.lapack_libraries

    cmake_opts['DCMAKE_INSTALL_RPATH'] = ";".join(rpaths)

    pwd = chdir(path)
    cmake('..', **cmake_opts)

    txt = 'serial' if serial else 'parallel'

    make('mfem_' + txt)
    make_install('mfem_' + txt)

    from shutil import copytree, rmtree

    print("copying mesh data for testing", "../data",
          cmake_opts['DCMAKE_INSTALL_PREFIX'])
    path = os.path.join(cmake_opts['DCMAKE_INSTALL_PREFIX'], "data")
    if os.path.exists(path):
        rmtree(path)
    copytree("../data", path)

    os.chdir(pwd)



from __future__ import print_function
import os
import sys
import numpy as np

from scipy.sparse import csr_matrix, coo_matrix


def run_test():
    import mfem.par as par
    from mfem.common.parcsr_extra import ToHypreParCSR, ToScipyCoo
    from mpi4py import MPI
    from mfem.common.mpi_debug import nicePrint

    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank

    device = mfem.Device('cuda')
    
    def print_hypre(M, txt):
        for i in range(num_proc):
            MPI.COMM_WORLD.Barrier()
            if myid == i:
                if myid == 0:
                    print(txt)
                    print('MyID: ', myid)
                else:
                    print('MyID: ', myid)
                print(ToScipyCoo(M))

    # make sample matrix
    row = np.array([0, 0, 1, 1])
    col = np.array([0, 3, 1, 2])
    data = np.array([4, 5, 7, 9])
    m = coo_matrix((data, (row, col)), shape=(2, 4))
    m = m.tocsr()
    m = m*(myid+1)

    M = ToHypreParCSR(m, assert_non_square_no_col_starts=False, verbose=True)

    print("memory class, memory type",
          mfem.GetHypreMemoryClass(), mfem.GetHypreMemoryType())

    print_hypre(M, 'matrix M')

    from mfem.common.chypre import CHypreVec

    r1 = np.array([0, 0, 1, 1])
    r2 = np.array([1, 1, 0, 0])
    vec1 = CHypreVec(r1, None)
    vec2 = CHypreVec(r2, None)

    if myid == 0:
        print("v1")
    v1 = (vec1-vec1*1j)
    v2 = (vec1+vec1*1j)

    nicePrint(v1.GlobalVector())
    nicePrint(v2.GlobalVector())
    nicePrint((v1+v2).GlobalVector())
    nicePrint((v1-v2).GlobalVector())

    if myid == 0:
        print("v1, v2")
    v1 = (vec1-vec2*1j)
    v2 = (vec1+vec2*1j)
    nicePrint(v1.GlobalVector())
    nicePrint(v2.GlobalVector())

    if myid == 0:
        print("v1+v2")
    nicePrint((v1+v2).GlobalVector())
    if myid == 0:
        print("v1-v2")
    nicePrint((v1-v2).GlobalVector())

    if myid == 0:
        print("3*v1")
    nicePrint(v1.GlobalVector())
    v1 *= 1j
    if myid == 0:
        print("1j*v1")
    nicePrint(v1.GlobalVector())
    if myid == 0:
        print("v1 dot v1")
    print(v1.dot(v1))
    if myid == 0:
        print("v1 *= 1+1j")
    v1 *= 1+1j
    nicePrint("v1", v1.GlobalVector())
    nicePrint("v2", v2.GlobalVector())
    #print(v1.dot(v1))
    #print(v1.dot(v2))
    #if mfem.is_HYPRE_USING_CUDA():
    #    v1.HypreRead()

    v1[0] += 3.
    if myid == 0:
        print("3*v1")
    #if mfem.is_HYPRE_USING_CUDA():
    #    v1.HypreRead()
        
    nicePrint(v1[0].GetDataArray())
    nicePrint(v1.GlobalVector())

if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-p':
        import mfem.par as mfem
        run_test()
    else:
        import mfem.ser as mfem



import sys
import numpy as np
from scipy.sparse import csr_matrix

if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True

    from mfem.common.mpi_debug import nicePrint as print

    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank

else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0



smat = csr_matrix([[1,2,3,], [0, 0, 1], [2, 0, 0]])
mmat = mfem.SparseMatrix(smat)

offset1 = mfem.intArray([0, 3, 6])
offset2 = mfem.intArray([0, 3, 6])

m = mfem.BlockMatrix(offset1, offset2)
m.SetBlock(0, 0, mmat)

print(m._offsets)
print(m._linked_mat)

m = mfem.BlockMatrix(offset1)
m.SetBlock(0, 1, mmat)
m.SetBlock(1, 0, mmat)
print(m._offsets)
print(m._linked_mat)

from mfem.common.sparse_utils import sparsemat_to_scipycsr
print(m.CreateMonolithic())
print(sparsemat_to_scipycsr(m.CreateMonolithic(), np.float64).todense())



'''
   MFEM example 5

   See c++ version in the MFEM library for more detail 
'''
import os
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp

# time.clock deprecated and removed in PY3.8
import time
try:
    clock = time.process_time
except AttributeError:
    clock = time.clock


def run(order=1,
        meshfile='',
        visualization=False,
        device='cpu',
        numba=False,
        pa=False):

    def pFunc_exact(x):
        xi = float(x[0])
        yi = float(x[1])
        zi = 0.0
        if len(x) == 3:
            zi = x[2]
        from numpy import sin, cos, exp
        return exp(xi)*sin(yi)*cos(zi)

    class uFunc_ex(mfem.VectorPyCoefficient):
        def EvalValue(self, x):
            xi = float(x[0])
            yi = float(x[1])
            zi = 0.0
            if len(x) == 3:
                zi = x[2]
            ret = [- exp(xi)*sin(yi)*cos(zi),
                   - exp(xi)*cos(yi)*cos(zi)]
            if len(x) == 3:
                ret.append(exp(xi)*sin(yi)*sin(zi))
            return ret

    class pFunc_ex(mfem.PyCoefficient):
        def EvalValue(self, x):
            return pFunc_exact(x)

    class fFunc(mfem.VectorPyCoefficient):
        def EvalValue(self, x):
            if len(x) == 3:
                return [0., 0., 0.]
            else:
                return [0., 0.]

    class gFunc(mfem.PyCoefficient):
        def EvalValue(self, x):
            if len(x) == 3:
                return -pFunc_exact(x)
            return 0.

    class f_natural(mfem.PyCoefficient):
        def EvalValue(self, x):
            return -pFunc_exact(x)

    device = mfem.Device(device)
    device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)

    dim = mesh.Dimension()

    ref_levels = int(np.floor(np.log(10000./mesh.GetNE())/np.log(2.)/dim))
    for x in range(ref_levels):
        mesh.UniformRefinement()

    hdiv_coll = mfem.RT_FECollection(order, dim)
    l2_coll = mfem.L2_FECollection(order, dim)

    R_space = mfem.FiniteElementSpace(mesh, hdiv_coll)
    W_space = mfem.FiniteElementSpace(mesh, l2_coll)

    dimR = R_space.GetVSize()
    dimW = W_space.GetVSize()

    print("***********************************************************")
    print("dim(R) = " + str(dimR))
    print("dim(W) = " + str(dimW))
    print("dim(R+W) = " + str(dimR+dimW))
    print("***********************************************************")

    block_offsets = intArray([0, dimR, dimW])
    block_offsets.PartialSum()

    k = mfem.ConstantCoefficient(1.0)

    fcoeff = fFunc(dim)
    fnatcoeff = f_natural()
    gcoeff = gFunc()
    ucoeff = uFunc_ex(dim)
    pcoeff = pFunc_ex()

    x = mfem.BlockVector(block_offsets)
    rhs = mfem.BlockVector(block_offsets)

    fform = mfem.LinearForm()
    fform.Update(R_space, rhs.GetBlock(0), 0)
    fform.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(fcoeff))
    fform.AddBoundaryIntegrator(
        mfem.VectorFEBoundaryFluxLFIntegrator(fnatcoeff))
    fform.Assemble()

    gform = mfem.LinearForm()
    gform.Update(W_space, rhs.GetBlock(1), 0)
    gform.AddDomainIntegrator(mfem.DomainLFIntegrator(gcoeff))
    gform.Assemble()

    mVarf = mfem.BilinearForm(R_space)
    bVarf = mfem.MixedBilinearForm(R_space, W_space)

    if pa:
        mVarf.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    mVarf.AddDomainIntegrator(mfem.VectorFEMassIntegrator(k))
    mVarf.Assemble()
    if not pa:
        mVarf.Finalize()

    if pa:
        bVarf.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    bVarf.AddDomainIntegrator(mfem.VectorFEDivergenceIntegrator())
    bVarf.Assemble()
    if not pa:
        bVarf.Finalize()

    darcyOp = mfem.BlockOperator(block_offsets)

    if pa:
        Bt = mfem.TransposeOperator(bVarf)

        darcyOp.SetBlock(0, 0, mVarf)
        darcyOp.SetBlock(0, 1, Bt, -1)
        darcyOp.SetBlock(1, 0, bVarf, -1)
    else:
        M = mVarf.SpMat()
        B = bVarf.SpMat()
        B *= -1
        if mfem.Device.IsEnabled():
            B = mfem.Transpose(B)
        Bt = mfem.TransposeOperator(B)

        darcyOp.SetBlock(0, 0, M)
        darcyOp.SetBlock(0, 1, Bt)
        darcyOp.SetBlock(1, 0, B)

    Md = mfem.Vector(mVarf.Height())
    darcyPrec = mfem.BlockDiagonalPreconditioner(block_offsets)

    if pa:
        mVarf.AssembleDiagonal(Md)
        Md_host = mfem.Vector(Md.HostRead(), mVarf.Height())
        invMd = mfem.Vector(mVarf.Height())

        for i in range(invMd.Size()):
            invMd[i] = 1.0/Md_host[i]

        BMBt_diag = mfem.Vector(bVarf.Height())
        bVarf.AssembleDiagonal_ADAt(invMd, BMBt_diag)
        ess_tdof_list = mfem.intArray()
        invM = mfem.OperatorJacobiSmoother(Md, ess_tdof_list)
        invS = mfem.OperatorJacobiSmoother(BMBt_diag, ess_tdof_list)
    else:
        MinvBt = mfem.Transpose(B)
        M.GetDiag(Md)

        for i in range(Md.Size()):
            MinvBt.ScaleRow(i, 1/Md[i])
        S = mfem.Mult(B, MinvBt)

        invM = mfem.DSmoother(M)
        invS = mfem.GSSmoother(S)

    invM.iterative_mode = False
    invS.iterative_mode = False

    darcyPrec.SetDiagonalBlock(0, invM)
    darcyPrec.SetDiagonalBlock(1, invS)

    maxIter = 1000
    rtol = 1e-6
    atol = 1e-10

    stime = clock()
    solver = mfem.MINRESSolver()
    solver.SetAbsTol(atol)
    solver.SetRelTol(rtol)
    solver.SetMaxIter(maxIter)
    solver.SetOperator(darcyOp)
    solver.SetPreconditioner(darcyPrec)
    solver.SetPrintLevel(1)
    x.Assign(0.0)
    solver.Mult(rhs, x)

    solve_time = clock() - stime

    if solver.GetConverged():
        print("MINRES converged in " + str(solver.GetNumIterations()) +
              " iterations with a residual norm of " + "{:g}".format(solver.GetFinalNorm()))
    else:
        print("MINRES did not converge in " + str(solver.GetNumIterations()) +
              " iterations. Residual norm is " + "{:g}".format(solver.GetFinalNorm()))
    print("MINRES solver took " + str(solve_time) + "s.")

    u = mfem.GridFunction()
    p = mfem.GridFunction()
    u.MakeRef(R_space, x.GetBlock(0), 0)
    p.MakeRef(W_space, x.GetBlock(1), 0)

    order_quad = max(2, 2*order+1)

    irs = [mfem.IntRules.Get(i, order_quad)
           for i in range(mfem.Geometry.NumGeom)]

    norm_p = mfem.ComputeLpNorm(2, pcoeff, mesh, irs)
    norm_u = mfem.ComputeLpNorm(2, ucoeff, mesh, irs)
    err_u = u.ComputeL2Error(ucoeff, irs)
    err_p = p.ComputeL2Error(pcoeff, irs)

    print("|| u_h - u_ex || / || u_ex || = " + "{:g}".format(err_u / norm_u))
    print("|| p_h - p_ex || / || p_ex || = " + "{:g}".format(err_p / norm_p))

    mesh.Print("ex5.mesh", 8)
    u.Save("sol_u.gf", 8)
    p.Save("sol_p.gf", 8)

    visit_dc = mfem.VisItDataCollection("Example5", mesh)
    visit_dc.RegisterField("velocity", u)
    visit_dc.RegisterField("pressure", p)
    visit_dc.Save()

    paraview_dc = mfem.ParaViewDataCollection("Example5", mesh)
    paraview_dc.SetPrefixPath("ParaView")
    paraview_dc.SetLevelsOfDetail(order)
    paraview_dc.SetCycle(0)
    paraview_dc.SetDataFormat(mfem.VTKFormat_BINARY)
    paraview_dc.SetHighOrderOutput(True)
    paraview_dc.SetTime(0.0)
    paraview_dc.RegisterField("velocity", u)
    paraview_dc.RegisterField("pressure", p)
    paraview_dc.Save()

    if visualization:
        sout_u = mfem.socketstream("localhost", 19916)
        sout_u.precision(8)
        sout_u << "solution\n" << mesh << u << "window_title 'Velocity'\n"
        sout_p = mfem.socketstream("localhost", 19916)
        sout_p.precision(8)
        sout_p << "solution\n" << mesh << p << "window_title 'Pressure'\n"


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex5 (Darcy Problem)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    args = parser.parse_args()
    parser.print_options(args)

    order = args.order
    meshfile = expanduser(
        join(dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly

    run(order=order,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa)



'''
   MFEM example 14

   See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np

ref_levels = -1
order = 1
sigma = -1.0
kappa = -1.0

if (kappa < 0):
    kappa = (order+1)**2.

meshfile = expanduser(
    join(dirname(__file__), '..', 'data', 'star.mesh'))
mesh = mfem.Mesh(meshfile, 1, 1)

dim = mesh.Dimension()

#   3. Refine the mesh to increase the resolution. In this example we do
#      'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
#      largest number that gives a final mesh with no more than 50,000
#      elements.
if ref_levels < 0:
    ref_levels = int(np.floor(np.log(50000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
    mesh.UniformRefinement()

if (mesh.NURBSext):
    mesh.SetCurvature(max(order, 1))

# 4. Define a finite element space on the mesh. Here we use discontinuous
#    finite elements of the specified order >= 0.
fec = mfem.DG_FECollection(order, dim)
fespace = mfem.FiniteElementSpace(mesh, fec)
print('Number of finite element unknowns: ' + str(fespace.GetVSize()))

# 5. Set up the linear form b(.) which corresponds to the right-hand side of
#    the FEM linear system.
b = mfem.LinearForm(fespace)
one = mfem.ConstantCoefficient(1.0)
zero = mfem.ConstantCoefficient(0.0)
b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
b.AddBdrFaceIntegrator(
    mfem.DGDirichletLFIntegrator(zero, one, sigma, kappa))
b.Assemble()

# 6. Define the solution vector x as a finite element grid function
#    corresponding to fespace. Initialize x with initial guess of zero.
x = mfem.GridFunction(fespace)
x.Assign(0.0)

# 7. Set up the bilinear form a(.,.) on the finite element space
#    corresponding to the Laplacian operator -Delta, by adding the Diffusion
#    domain integrator and the interior and boundary DG face integrators.
#    Note that boundary conditions are imposed weakly in the form, so there
#    is no need for dof elimination. After assembly and finalizing we
#    extract the corresponding sparse matrix A.
a = mfem.BilinearForm(fespace)
a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))

a.AddInteriorFaceIntegrator(mfem.DGDiffusionIntegrator(one, sigma, kappa))
a.AddBdrFaceIntegrator(mfem.DGDiffusionIntegrator(one, sigma, kappa))
a.Assemble()
a.Finalize()
A = a.SpMat()

# 8. Define a simple symmetric Gauss-Seidel preconditioner and use it to
#    solve the system Ax=b with PCG in the symmetric case, and GMRES in the
#    non-symmetric one.
M = mfem.GSSmoother(A)
if (sigma == -1.0):
    mfem.PCG(A, M, b, x, 1, 500, 1e-12, 0.0)
else:
    mfem.GMRES(A, M, b, x, 1, 500, 10, 1e-12, 0.0)

# 9. Save the refined mesh and the solution. This output can be viewed later
#    using GLVis: "glvis -m refined.mesh -g sol.gf".
mesh.Print('refined.mesh', 8)
x.Save('sol.gf', 8)



'''
   PyMFEM example 36

   See c++ version in the MFEM library for more detail

   Sample runs:  
      mpirun -np 4 python ex36.py -o 2
      mpirun -np 4 python ex36.py -o 2 -r 4


'''
import os
from os.path import expanduser, join

from numpy import sqrt, log, exp
import numpy as np
from numba import njit
from numba.types import float64


import mfem.ser as mfem
from mfem.ser import intArray, doubleArray


visualization = True


def run(refs=3,
        order=1,
        max_it=10,
        tol=1e-5,
        alpha=1.0):

    # 2. Read the mesh from the mesh file.
    mesh_file = "../data/disc-nurbs.mesh"
    mesh_file = expanduser(
        join(os.path.dirname(__file__), *mesh_file.split("/")))
    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()

    # 3. Postprocess the mesh.
    # 3A. Refine the mesh to increase the resolution.
    for l in range(refs):
        mesh.UniformRefinement()

    # 3B. Interpolate the geometry after refinement to control geometry error.
    # NOTE: Minimum second-order interpolation is used to improve the accuracy.
    curvature_order = max((order, 2))
    mesh.SetCurvature(curvature_order)

    # 3C. Rescale the domain to a unit circle (radius = 1).
    nodes = mesh.GetNodes()
    scale = 2*sqrt(2)
    nodes /= scale

    # 4. Define the necessary finite element spaces on the mesh.
    H1fec = mfem.H1_FECollection(order+1, dim)
    H1fes = mfem.FiniteElementSpace(mesh, H1fec)

    L2fec = mfem.L2_FECollection(order-1, dim)
    L2fes = mfem.FiniteElementSpace(mesh, L2fec)

    print("Number of H1 finite element unknowns: " +
          str(H1fes.GetTrueVSize()))
    print("Number of L2 finite element unknowns: " +
          str(L2fes.GetTrueVSize()))

    offsets = mfem.intArray((0, H1fes.GetVSize(), L2fes.GetVSize()))
    offsets.PartialSum()

    x = mfem.BlockVector(offsets)
    rhs = mfem.BlockVector(offsets)
    x.Assign(0.0)
    rhs.Assign(0.0)

    # 5. Determine the list of true (i.e. conforming) essential boundary dofs.
    if mesh.bdr_attributes.Size() > 0:
        ess_bdr = mfem.intArray([1]*mesh.bdr_attributes.Max())
    else:
        ess_bdr = mfem.intArray()

    # 6. Define an initial guess for the solution.
    @njit(float64(float64[:]))
    def IC_func(x):
        r0 = 1.0
        rr = 0.0
        for i in range(len(x)):
            rr += x[i]*x[i]
        return r0*r0 - rr

    one = mfem.ConstantCoefficient(1.0)
    zero = mfem.ConstantCoefficient(0.0)

    # 7. Define the solution vectors as a finite element grid functions
    #    corresponding to the fespaces.
    u_gf = mfem.GridFunction()
    delta_psi_gf = mfem.GridFunction()
    u_gf.MakeRef(H1fes, x, offsets[0])
    delta_psi_gf.MakeRef(L2fes, x, offsets[1])
    delta_psi_gf.Assign(0.0)

    u_old_gf = mfem.GridFunction(H1fes)
    psi_old_gf = mfem.GridFunction(L2fes)
    psi_gf = mfem.GridFunction(L2fes)
    u_old_gf.Assign(0.0)
    psi_old_gf.Assign(0.0)

    # 8. Define the function coefficients for the solution and use them to
    #    initialize the initial guess
    @mfem.jit.scalar
    def exact_coef(ptx):
        return exact_solution_obstacle(ptx[0], ptx[1])

    @mfem.jit.vector(shape=(dim, ))
    def exact_grad_coef(ptx):
        return exact_solution_gradient_obstacle(ptx[0], ptx[1])

    @mfem.jit.scalar
    def IC_coef(ptx):
        return IC_func(ptx)
    f = mfem.ConstantCoefficient(0.0)

    @mfem.jit.scalar
    def obstacle(ptx):
        return spherical_obstacle(ptx[0], ptx[1])

    u_gf.ProjectCoefficient(IC_coef)
    u_old_gf.GetDataArray()[:] = u_gf.GetDataArray()

    # 9. Initialize the slack variable ψₕ = ln(uₕ)
    c_u_gf = mfem.GridFunctionCoefficient(u_gf)

    @mfem.jit.scalar(dependency=(c_u_gf, obstacle))
    def ln_u(ptx, u, obstacle):
        min_val = -36.
        val = u - obstacle
        return max((min_val, log(val)))

    psi_gf.ProjectCoefficient(ln_u)
    psi_old_gf.GetDataArray()[:] = psi_gf.GetDataArray()

    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)

    # Note: we define exp_psi outside the iteration loop to avoid
    #       numba-jit during the loop.
    #       GridFunction referred from c_psi_gf is updated during
    #       the iteration.
    c_psi_gf = mfem.GridFunctionCoefficient()

    @mfem.jit.scalar(dependency=(c_psi_gf, ))
    def exp_psi(ptx, val):
        min_val = 0
        max_val = 1e36
        return min(max_val, max(min_val, exp(val)))

    # 10. Iterate
    total_iterations = 0
    increment_u = 0.1
    for k in range(max_it):
        u_tmp = mfem.GridFunction(H1fes)
        u_tmp.GetDataArray()[:] = u_old_gf.GetDataArray()

        print("\nOUTER ITERATION " + str(k+1))

        for j in range(10):
            total_iterations += 1

            alpha_cf = mfem.ConstantCoefficient(alpha)

            b0 = mfem.LinearForm()
            b1 = mfem.LinearForm()
            b0.Update(H1fes, rhs.GetBlock(0), 0)
            b1.Update(L2fes, rhs.GetBlock(1), 0)

            c_psi_gf.SetGridFunction(psi_gf)
            neg_exp_psi = mfem.ProductCoefficient(-1.0, exp_psi)
            grad_u_old = mfem.GradientGridFunctionCoefficient(u_old_gf)
            alpha_f = mfem.ProductCoefficient(alpha, f)
            psi_cf = mfem.GridFunctionCoefficient(psi_gf)
            psi_old_cf = mfem.GridFunctionCoefficient(psi_old_gf)
            psi_old_minus_psi = mfem.SumCoefficient(
                psi_old_cf, psi_cf, 1.0, -1.0)

            b0.AddDomainIntegrator(mfem.DomainLFIntegrator(alpha_f))
            b0.AddDomainIntegrator(mfem.DomainLFIntegrator(psi_old_minus_psi))
            b0.Assemble()

            b1.AddDomainIntegrator(mfem.DomainLFIntegrator(exp_psi))
            b1.AddDomainIntegrator(mfem.DomainLFIntegrator(obstacle))
            b1.Assemble()

            a00 = mfem.BilinearForm(H1fes)
            a00.SetDiagonalPolicy(mfem.Operator.DIAG_ONE)
            a00.AddDomainIntegrator(mfem.DiffusionIntegrator(alpha_cf))
            a00.Assemble()
            a00.EliminateEssentialBC(ess_bdr, x.GetBlock(0), rhs.GetBlock(0),
                                     mfem.Operator.DIAG_ONE)
            a00.Finalize()
            A00 = a00.SpMat()

            a10 = mfem.MixedBilinearForm(H1fes, L2fes)
            a10.AddDomainIntegrator(mfem.MixedScalarMassIntegrator())
            a10.Assemble()
            a10.EliminateTrialDofs(ess_bdr, x.GetBlock(0), rhs.GetBlock(1))
            a10.Finalize()
            A10 = a10.SpMat()

            A01 = mfem.Transpose(A10)

            a11 = mfem.BilinearForm(L2fes)
            a11.AddDomainIntegrator(mfem.MassIntegrator(neg_exp_psi))
            # NOTE: Shift the spectrum of the Hessian matrix for additional
            #       stability (Quasi-Newton).
            eps_cf = mfem.ConstantCoefficient(-1e-6)
            if (order == 1):
                # NOTE: ∇ₕuₕ = 0 for constant functions.
                #       Therefore, we use the mass matrix to shift the spectrum
                a11.AddDomainIntegrator(mfem.MassIntegrator(eps_cf))
            else:
                a11.AddDomainIntegrator(mfem.DiffusionIntegrator(eps_cf))

            a11.Assemble()
            a11.Finalize()
            A11 = a11.SpMat()

            A = mfem.BlockOperator(offsets)
            A.SetBlock(0, 0, A00)
            A.SetBlock(1, 0, A10)
            A.SetBlock(0, 1, A01)
            A.SetBlock(1, 1, A11)

            prec = mfem.BlockDiagonalPreconditioner(offsets)
            gs1 = mfem.GSSmoother(A00)
            gs2 = mfem.GSSmoother(A11)
            prec.SetDiagonalBlock(0, gs1)
            prec.SetDiagonalBlock(1, gs2)
            prec.owns_blocks = 0

            mfem.GMRES(A, prec, rhs, x, 0, 10000, 500, 1e-12, 0.0)

            u_gf.MakeRef(H1fes, x.GetBlock(0), 0)
            delta_psi_gf.MakeRef(L2fes, x.GetBlock(1), 0)

            u_tmp -= u_gf
            Newton_update_size = u_tmp.ComputeL2Error(zero)
            u_tmp.Assign(u_gf)
            # Note: to copy GridFunction data, use either Assign or
            #       numpy representation.
            #    u_tmp.Assign(u_gf)
            #    u_tmp.GetDataArray()[:] = u_gf.GetDataArray()

            gamma = 1.0
            delta_psi_gf *= gamma
            psi_gf += delta_psi_gf

            if visualization:
                sol_sock << "solution\n" << mesh << u_gf
                sol_sock << "window_title 'Discrete solution'"
                sol_sock.flush()
            print("Newton_update_size = " +
                  "{:g}".format(Newton_update_size))

            del A01

            if Newton_update_size < increment_u:
                break

        u_tmp.GetDataArray()[:] = u_gf.GetDataArray()
        u_tmp -= u_old_gf
        increment_u = u_tmp.ComputeL2Error(zero)

        print("Number of Newton iterations = " + str(j+1))
        print("Increment (|| uₕ - uₕ_prvs||) = " + "{:g}".format(increment_u))

        u_old_gf.GetDataArray()[:] = u_gf.GetDataArray()
        psi_old_gf.GetDataArray()[:] = psi_gf.GetDataArray()

        if increment_u < tol or k == max_it-1:
            break

        H1_error = u_gf.ComputeH1Error(exact_coef, exact_grad_coef)
        print("H1-error  (|| u - uₕᵏ||)       = " + "{:g}".format(H1_error))

    print("Outer iterations: " + str(k+1))
    print(" Total iterations: " + str(total_iterations))
    print(" Total dofs:       " + str(H1fes.GetTrueVSize() + L2fes.GetTrueVSize()))

    # 11. Exact solution.
    if visualization:
        err_sock = mfem.socketstream("localhost", 19916)
        err_sock.precision(8)

        error_gf = mfem.GridFunction(H1fes)
        error_gf.ProjectCoefficient(exact_coef)
        error_gf -= u_gf

        err_sock << "solution\n" << mesh << error_gf << "window_title 'Error'"
        err_sock.flush()

    L2_error = u_gf.ComputeL2Error(exact_coef)
    H1_error = u_gf.ComputeH1Error(exact_coef, exact_grad_coef)

    c_psi_gf = mfem.GridFunctionCoefficient(psi_gf)

    @mfem.jit.scalar(dependency=(c_psi_gf, obstacle))
    def u_alt_cf(ptx, val, obstacle):
        min_val = 0
        max_val = 1e36
        return min(max_val, max(min_val, exp(val)+obstacle))

    u_alt_gf = mfem.GridFunction(L2fes)
    u_alt_gf.ProjectCoefficient(u_alt_cf)
    L2_error_alt = u_alt_gf.ComputeL2Error(exact_coef)

    print("\n Final L2-error (|| u - uₕ||)          = " +
          "{:g}".format(L2_error))
    print(" Final H1-error (|| u - uₕ||)          = " +
          "{:g}".format(H1_error))
    print(" Final L2-error (|| u - ϕ - exp(ψₕ)||) = " +
          "{:g}".format(L2_error_alt))


@njit(float64(float64, float64))
def spherical_obstacle(x, y):
    r = sqrt(x*x + y*y)
    r0 = 0.5
    beta = 0.9

    b = r0*beta
    tmp = sqrt(r0*r0 - b*b)
    B = tmp + b*b/tmp
    C = -b/tmp

    if r > b:
        return B + r * C
    else:
        return sqrt(r0*r0 - r*r)


@njit(float64(float64, float64))
def exact_solution_obstacle(x, y):
    r = sqrt(x*x + y*y)
    r0 = 0.5
    a = 0.348982574111686
    A = -0.340129705945858

    if r > a:
        return A * log(r)
    else:
        return sqrt(r0*r0-r*r)


@njit(float64[:](float64, float64))
def exact_solution_gradient_obstacle(x, y):
    r = sqrt(x*x + y*y)
    r0 = 0.5
    a = 0.348982574111686
    A = -0.340129705945858

    grad = np.zeros(2, dtype=np.float64)
    if r > a:
        grad[0] = A * x / (r*r)
        grad[1] = A * y / (r*r)
    else:
        grad[0] = - x / sqrt(r0*r0 - r*r)
        grad[1] = - y / sqrt(r0*r0 - r*r)
    return grad


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex36 (Obstacle problem)')

    parser.add_argument("-o", "--order",
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree).")
    parser.add_argument('-r', '--refs',
                        action='store', default=3, type=int,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument('-mi', '--max-it',
                        action='store', default=10, type=int,
                        help="Maximum number of iterations")
    help = "\n".join(("Stopping criteria based on the difference between",
                      "successive solution updates"))
    parser.add_argument('-tol', '--tol',
                        action='store', default=1e-5, type=float,
                        help=help)
    parser.add_argument('-step', '--step',
                        action='store', default=1.0, type=float,
                        help="Step size alpha")
    parser.add_argument("-no-vis", "--no-visualization",
                        action='store_true', default=False,
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    parser.print_options(args)

    globals()["visualization"] = not args.no_visualization

    run(refs=args.refs,
        order=args.order,
        max_it=args.max_it,
        tol=args.tol,
        alpha=args.step)



'''
   MFEM example 5p

   See c++ version in the MFEM library for more detail 
'''
import time
from numpy import sin, cos, exp
import numpy as np
from mfem import path
import mfem.par as mfem
from mfem.par import intArray
from mfem.common.mpi_debug import nicePrint

from os.path import expanduser, join, dirname

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)
verbose = (myid == 0)


# time.clock deprecated and removed in PY3.8
try:
    clock = time.process_time
except AttributeError:
    clock = time.clock


def run(order=1,
        refine=-1,
        meshfile='',
        visualization=False,
        par_format=False,
        device='cpu',
        numba=False,
        pa=False):

    def pFunc_exact(x):
        xi = float(x[0])
        yi = float(x[1])
        zi = 0.0
        if len(x) == 3:
            zi = x[2]
        from numpy import sin, cos, exp
        return exp(xi)*sin(yi)*cos(zi)

    class uFunc_ex(mfem.VectorPyCoefficient):
        def EvalValue(self, x):
            xi = float(x[0])
            yi = float(x[1])
            zi = 0.0
            if len(x) == 3:
                zi = x[2]
            ret = [- exp(xi)*sin(yi)*cos(zi),
                   - exp(xi)*cos(yi)*cos(zi)]
            if len(x) == 3:
                ret.append(exp(xi)*sin(yi)*sin(zi))
            return ret

    class pFunc_ex(mfem.PyCoefficient):
        def EvalValue(self, x):
            return pFunc_exact(x)

    class fFunc(mfem.VectorPyCoefficient):
        def EvalValue(self, x):
            if len(x) == 3:
                return [0., 0., 0.]
            else:
                return [0., 0.]

    class gFunc(mfem.PyCoefficient):
        def EvalValue(self, x):
            if len(x) == 3:
                return -pFunc_exact(x)
            else:
                return 0.0

    class f_natural(mfem.PyCoefficient):
        def EvalValue(self, x):
            return -pFunc_exact(x)

    device = mfem.Device(device)
    if myid == 0:
        device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    if refine == -1:
        ref_levels = int(np.floor(np.log(10000./mesh.GetNE())/np.log(2.)/dim))
        for x in range(ref_levels):
            mesh.UniformRefinement()
    for i in range(refine):
        mesh.UniformRefinement()

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    par_ref_levels = 2
    for l in range(par_ref_levels):
        pmesh.UniformRefinement()

    hdiv_coll = mfem.RT_FECollection(order, dim)
    l2_coll = mfem.L2_FECollection(order, dim)

    R_space = mfem.ParFiniteElementSpace(pmesh, hdiv_coll)
    W_space = mfem.ParFiniteElementSpace(pmesh, l2_coll)

    dimR = R_space.GlobalTrueVSize()
    dimW = W_space.GlobalTrueVSize()

    if verbose:
        print("***********************************************************")
        print("dim(R) = " + str(dimR))
        print("dim(W) = " + str(dimW))
        print("dim(R+W) = " + str(dimR+dimW))
        print("***********************************************************")

    block_offsets = intArray([0, R_space.GetVSize(), W_space.GetVSize()])
    block_offsets.PartialSum()
    block_trueOffsets = intArray([0, R_space.TrueVSize(), W_space.TrueVSize()])
    block_trueOffsets.PartialSum()

    k = mfem.ConstantCoefficient(1.0)

    fcoeff = fFunc(dim)
    fnatcoeff = f_natural()
    gcoeff = gFunc()
    ucoeff = uFunc_ex(dim)
    pcoeff = pFunc_ex()

    x = mfem.BlockVector(block_offsets)

    rhs = mfem.BlockVector(block_offsets)
    trueX = mfem.BlockVector(block_trueOffsets)

    trueRhs = mfem.BlockVector(block_trueOffsets)
    trueRhs.Assign(0.0)

    fform = mfem.ParLinearForm()
    fform.Update(R_space, rhs.GetBlock(0), 0)
    fform.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(fcoeff))
    fform.AddBoundaryIntegrator(
        mfem.VectorFEBoundaryFluxLFIntegrator(fnatcoeff))
    fform.Assemble()
    fform.ParallelAssemble(trueRhs.GetBlock(0))

    gform = mfem.ParLinearForm()
    gform.Update(W_space, rhs.GetBlock(1), 0)
    gform.AddDomainIntegrator(mfem.DomainLFIntegrator(gcoeff))
    gform.Assemble()
    gform.ParallelAssemble(trueRhs.GetBlock(1))

    mVarf = mfem.ParBilinearForm(R_space)
    bVarf = mfem.ParMixedBilinearForm(R_space, W_space)

    if pa:
        mVarf.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    mVarf.AddDomainIntegrator(mfem.VectorFEMassIntegrator(k))
    mVarf.Assemble()
    if not pa:
        mVarf.Finalize()

    if pa:
        bVarf.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    bVarf.AddDomainIntegrator(mfem.VectorFEDivergenceIntegrator())
    bVarf.Assemble()
    if not pa:
        bVarf.Finalize()

    darcyOp = mfem.BlockOperator(block_trueOffsets)

    if pa:
        opM = mfem.OperatorPtr()
        opB = mfem.OperatorPtr()

        empty_tdof_list = mfem.intArray()
        mVarf.FormSystemMatrix(empty_tdof_list, opM)
        bVarf.FormRectangularSystemMatrix(
            empty_tdof_list, empty_tdof_list, opB)
        Bt = mfem.TransposeOperator(opB.Ptr())

        darcyOp.SetBlock(0, 0, opM.Ptr())
        darcyOp.SetBlock(0, 1, Bt, -1.0)
        darcyOp.SetBlock(1, 0, opB.Ptr(), -1.0)
    else:
        M = mVarf.ParallelAssemble()
        B = bVarf.ParallelAssemble()
        B *= -1
        Bt = mfem.TransposeOperator(B)

        darcyOp.SetBlock(0, 0, M)
        darcyOp.SetBlock(0, 1, Bt)
        darcyOp.SetBlock(1, 0, B)

    if pa:
        Md_PA = mfem.Vector(R_space.GetTrueVSize())
        mVarf.AssembleDiagonal(Md_PA)
        Md_host = mfem.Vector(Md_PA.HostRead(), R_space.GetTrueVSize())
        invMd = mfem.Vector(1.0 / Md_host.GetDataArray())

        BMBt_diag = mfem.Vector(bVarf.Height())
        bVarf.AssembleDiagonal_ADAt(invMd, BMBt_diag)
        ess_tdof_list = mfem.intArray()
        invM = mfem.OperatorJacobiSmoother(Md_PA, ess_tdof_list)
        invS = mfem.OperatorJacobiSmoother(BMBt_diag, ess_tdof_list)
    else:
        Md = mfem.HypreParVector(MPI.COMM_WORLD, M.GetGlobalNumRows(),
                                 M.GetRowStarts())
        M.GetDiag(Md)

        MinvBt = B.Transpose()
        MinvBt.InvScaleRows(Md)
        S = mfem.hypre.ParMult(B, MinvBt)
        invM = mfem.HypreDiagScale(M)
        invS = mfem.HypreBoomerAMG(S)

    invM.iterative_mode = False
    invS.iterative_mode = False

    darcyPr = mfem.BlockDiagonalPreconditioner(block_trueOffsets)
    darcyPr.SetDiagonalBlock(0, invM)
    darcyPr.SetDiagonalBlock(1, invS)

    maxIter = 1000 if pa else 500
    rtol = 1e-6
    atol = 1e-10

    stime = clock()
    solver = mfem.MINRESSolver(MPI.COMM_WORLD)

    solver.SetAbsTol(atol)
    solver.SetRelTol(rtol)
    solver.SetMaxIter(maxIter)
    solver.SetOperator(darcyOp)
    solver.SetPreconditioner(darcyPr)
    solver.SetPrintLevel(1)
    trueX.Assign(0.0)
    solver.Mult(trueRhs, trueX)
    if device.IsEnabled():
        trueX.HostRead()

    solve_time = clock() - stime

    if verbose:
        if solver.GetConverged():
            print("MINRES converged in " + str(solver.GetNumIterations()) +
                  " iterations with a residual norm of " +
                  "{:g}".format(solver.GetFinalNorm()))
        else:
            print("MINRES did not converge in " + str(solver.GetNumIterations()) +
                  " iterations. Residual norm is " +
                  "{:g}".format(solver.GetFinalNorm()))
        print("MINRES solver took " + "{:g}".format(solve_time) + "s.")

    u = mfem.ParGridFunction()
    p = mfem.ParGridFunction()
    u.MakeRef(R_space, x.GetBlock(0), 0)
    p.MakeRef(W_space, x.GetBlock(1), 0)
    u.Distribute(trueX.GetBlock(0))
    p.Distribute(trueX.GetBlock(1))

    order_quad = max(2, 2*order+1)
    irs = [mfem.IntRules.Get(i, order_quad)
           for i in range(mfem.Geometry.NumGeom)]

    err_u = u.ComputeL2Error(ucoeff, irs)
    norm_u = mfem.ComputeGlobalLpNorm(2, ucoeff, pmesh, irs)
    err_p = p.ComputeL2Error(pcoeff, irs)
    norm_p = mfem.ComputeGlobalLpNorm(2, pcoeff, pmesh, irs)

    if verbose:
        print("|| u_h - u_ex || / || u_ex || = " +
              "{:g}".format(err_u / norm_u))
        print("|| p_h - p_ex || / || p_ex || = " +
              "{:g}".format(err_p / norm_p))

    #  Save the refined mesh and the solution in parallel. This output can be
    #  viewed later using GLVis: "glvis -np <np> -m mesh -g sol_*".
    pmesh.Print('mesh.' + smyid, 8)
    u.Save('sol_u.' + smyid, 8)
    p.Save('sol_p.' + smyid, 8)

    # Save data in the Visit format
    visit_dc = mfem.VisItDataCollection("Example5-Parallel", pmesh)
    visit_dc.RegisterField("velocity", u)
    visit_dc.RegisterField("pressure", p)
    visit_format = (mfem.DataCollection.PARALLEL_FORMAT if par_format else
                    mfem.DataCollection.SERIAL_FORMAT)
    visit_dc.SetFormat(visit_format)
    visit_dc.Save()

    # Save data in the ParaVier format
    paraview_dc = mfem.ParaViewDataCollection("Example5P", pmesh)
    paraview_dc.SetPrefixPath("ParaView")
    paraview_dc.SetLevelsOfDetail(order)
    paraview_dc.SetCycle(0)
    paraview_dc.SetDataFormat(mfem.VTKFormat_BINARY)
    paraview_dc.SetHighOrderOutput(True)
    paraview_dc.SetTime(0.0)
    paraview_dc.RegisterField("velocity", u)
    paraview_dc.RegisterField("pressure", p)
    paraview_dc.Save()

    if visualization:
        u_sock = mfem.socketstream("localhost", 19916)
        u_sock << "parallel " << num_procs << " " << myid << "\n"
        u_sock.precision(8)
        u_sock << "solution\n" << pmesh << u << "window_title 'Velocity'\n"
        MPI.COMM_WORLD.Barrier()
        p_sock = mfem.socketstream("localhost", 19916)
        p_sock << "parallel " << num_procs << " " << myid << "\n"
        p_sock.precision(8)
        p_sock << "solution\n" << pmesh << p << "window_title 'Pressure'\n"


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex5 (Darcy Problem)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-r', '--refine',
                        action='store', type=int, default=-1,
                        help='Number of times to refine the mesh uniformly')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-pf', '--parallel-format',
                        action='store_true',
                        help="Use parallel format for Visit output")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    order = args.order
    meshfile = expanduser(
        join(dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly

    run(order=order,
        meshfile=meshfile,
        visualization=visualization,
        refine=args.refine,
        par_format=args.parallel_format,
        device=device,
        pa=pa)



'''
   MFEM example 6
      This is a version of Example 1 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np

order = 1
path = dirname((__file__))
meshfile = expanduser(join(path, '..', 'data', 'star.mesh'))

mesh = mfem.Mesh(meshfile, 1, 1)

dim = mesh.Dimension()
sdim = mesh.SpaceDimension()

# 3. Since a NURBS mesh can currently only be refined uniformly, we need to
#    convert it to a piecewise-polynomial curved mesh. First we refine the
#    NURBS mesh a bit more and then project the curvature to quadratic Nodes.
if (mesh.NURBSext):
    for i in range(2):
        mesh.UniformRefinement()
    mesh.SetCurvature(2)

# 4. Define a finite element space on the mesh. The polynomial order is
#      one (linear) by default, but this can be changed on the command line.
fec = mfem.H1_FECollection(order, dim)
fespace = mfem.FiniteElementSpace(mesh, fec)

# 5. As in Example 1, we set up bilinear and linear forms corresponding to
#    the Laplace problem -\Delta u = 1. We don't assemble the discrete
#    problem yet, this will be done in the main loop.
a = mfem.BilinearForm(fespace)
b = mfem.LinearForm(fespace)

one = mfem.ConstantCoefficient(1.0)
zero = mfem.ConstantCoefficient(0.0)

integ = mfem.DiffusionIntegrator(one)
a.AddDomainIntegrator(integ)
b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))

# 6. The solution vector x and the associated finite element grid function
#    will be maintained over the AMR iterations. We initialize it to zero.
x = mfem.GridFunction(fespace)
x.Assign(0.0)

#  7. All boundary attributes will be used for essential (Dirichlet) BC.
ess_bdr = intArray(mesh.bdr_attributes.Max())
ess_bdr.Assign(1)


# 9. Set up an error estimator. Here we use the Zienkiewicz-Zhu estimator
#    that uses the ComputeElementFlux method of the DiffusionIntegrator to
#    recover a smoothed flux (gradient) that is subtracted from the element
#    flux to get an error indicator. We need to supply the space for the
#    smoothed flux: an (H1)^sdim (i.e., vector-valued) space is used here.
flux_fespace = mfem.FiniteElementSpace(mesh, fec, sdim)

# own_flux_fes = False indicate flux_fespace is passed by reference
# this is actually default action, but for the sake of explanaiton
# it is explicitly set. If you want to pass pointer use own_flux_fes = True
estimator = mfem.ZienkiewiczZhuEstimator(integ, x, flux_fespace,
                                         own_flux_fes=False)
estimator.SetAnisotropic()

# 10. A refiner selects and refines elements based on a refinement strategy.
#     The strategy here is to refine elements with errors larger than a
#     fraction of the maximum element error. Other strategies are possible.
#     The refiner will call the given error estimator.
refiner = mfem.ThresholdRefiner(estimator)
refiner.SetTotalErrorFraction(0.7)


# 11. The main AMR loop. In each iteration we solve the problem on the
#     current mesh, visualize the solution, and refine the mesh.
max_dofs = 50000
it = 0
while True:
    cdofs = fespace.GetTrueVSize()
    print("AMR iteration " + str(it))
    print("Number of unknowns: " + str(cdofs))

    # 12. Assemble the stiffness matrix and the right-hand side.
    a.Assemble()
    b.Assemble()

    # 13. Set Dirichlet boundary values in the GridFunction x.
    #     Determine the list of Dirichlet true DOFs in the linear system.
    ess_tdof_list = intArray()
    x.ProjectBdrCoefficient(zero, ess_bdr)
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 14. Create the linear system: eliminate boundary conditions, constrain
    #     hanging nodes and possibly apply other transformations. The system
    #     will be solved for true (unconstrained) DOFs only.
    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    copy_interior = 1
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B, copy_interior)

    #  15. Define a simple symmetric Gauss-Seidel preconditioner and use it to
    #     solve the linear system with PCG.
    AA = mfem.OperatorHandle2SparseMatrix(A)
    M = mfem.GSSmoother(AA)
    mfem.PCG(AA, M, B, X, 3, 200, 1e-12, 0.0)

    # 16. After solving the linear system, reconstruct the solution as a
    #     finite element GridFunction. Constrained nodes are interpolated
    #     from true DOFs (it may therefore happen that x.Size() >= X.Size()).
    a.RecoverFEMSolution(X, b, x)
    if (cdofs > max_dofs):
        print("Reached the maximum number of dofs. Stop.")
        break
    # 18. Call the refiner to modify the mesh. The refiner calls the error
    #     estimator to obtain element errors, then it selects elements to be
    #     refined and finally it modifies the mesh. The Stop() method can be
    #     used to determine if a stopping criterion was met.
    refiner.Apply(mesh)
    if (refiner.Stop()):
        print("Stopping criterion satisfied. Stop.")
        break
    # 19. Update the space to reflect the new state of the mesh. Also,
    #     interpolate the solution x so that it lies in the new space but
    #     represents the same function. This saves solver iterations later
    #     since we'll have a good initial guess of x in the next step.
    #     Internally, FiniteElementSpace::Update() calculates an
    #     interpolation matrix which is then used by GridFunction::Update().
    fespace.Update()
    x.Update()

    # 20. Inform also the bilinear and linear forms that the space has
    #     changed.
    a.Update()
    b.Update()

    it = it + 1



'''
   MFEM example 0 (converted from ex0.cpp)

   See c++ version in the MFEM library for more detail

   How to run:
      python <arguments>

   Example of arguments:
      ex1.py -m star.mesh
      ex1.py -m fichera.mesh -o 2

   Description: This example code demonstrates the most basic usage of MFEM to
                define a simple finite element discretization of the Laplace
                problem -Delta u = 1 with zero Dirichlet boundary conditions.
                General 2D/3D mesh files and finite element polynomial degrees
                can be specified by command line options.

'''
import os
from os.path import expanduser, join
import numpy as np

import mfem.par as mfem
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)


def run(order=1, meshfile=''):
    '''
    run ex0
    '''

    #  2. Read the mesh from the given mesh file, and refine once uniformly.
    serial_mesh = mfem.Mesh(meshfile)
    mesh = mfem.ParMesh(MPI.COMM_WORLD, serial_mesh)
    mesh.UniformRefinement()

    # 3. Define a finite element space on the mesh. Here we use H1 continuous
    #    high-order Lagrange finite elements of the given order.
    fec = mfem.H1_FECollection(order,  mesh.Dimension())
    fespace = mfem.ParFiniteElementSpace(mesh, fec)
    gtdof = fespace.GlobalTrueVSize()

    if myid == 0:
        print('Number of finite element unknowns: ' + str(gtdof))

    # 4. Extract the list of all the boundary DOFs. These will be marked as
    #    Dirichlet in order to enforce zero boundary conditions.
    boundary_dofs = mfem.intArray()
    fespace.GetBoundaryTrueDofs(boundary_dofs)

    # 5. Define the solution x as a finite element grid function in fespace. Set
    #    the initial guess to zero, which also sets the boundary conditions.
    x = mfem.ParGridFunction(fespace)
    x.Assign(0.0)

    # 6. Set up the linear form b(.) corresponding to the right-hand side.
    one = mfem.ConstantCoefficient(1.0)
    b = mfem.ParLinearForm(fespace)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
    b.Assemble()

    # 7. Set up the bilinear form a(.,.) corresponding to the -Delta operator.
    a = mfem.ParBilinearForm(fespace)
    a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
    a.Assemble()

    # 8. Form the linear system A X = B. This includes eliminating boundary
    #    conditions, applying AMR constraints, and other transformations.
    A = mfem.HypreParMatrix()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(boundary_dofs, x, b, A, X, B)

    # 9. Solve the system using PCG with symmetric Gauss-Seidel preconditioner.
    M = mfem.HypreBoomerAMG(A)
    cg = mfem.CGSolver(MPI.COMM_WORLD)
    cg.SetRelTol(1e-12)
    cg.SetMaxIter(2000)
    cg.SetPrintLevel(1)
    cg.SetPreconditioner(M)
    cg.SetOperator(A)
    cg.Mult(B, X)

    # 10. Recover the solution x as a grid function and save to file. The output
    #     can be viewed using GLVis as follows: "glvis -m mesh.mesh -g sol.gf"
    a.RecoverFEMSolution(X, b, x)
    x.Save('sol.'+smyid)
    mesh.Print('mesh.'+smyid)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex1 (Laplace Problem)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")

    args = parser.parse_args()
    parser.print_options(args)

    order = args.order
    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    run(order=order,
        meshfile=meshfile)
